import datetime
from typing import Callable

IsHolidayFuncType = Callable[[datetime.datetime | datetime.date], bool]  # noqa: E501


def is_saturday_or_sunday(
    date: datetime.datetime | datetime.date,
) -> bool:
    """Check if the given date is a Saturday or Sunday."

    Args:
        date (datetime.datetime | datetime.date): Date to check.

    Returns:
        bool: True if the date is a Saturday or Sunday, False otherwise.
    """
    return date.weekday() in {5, 6}


def is_new_year_day(
    date: datetime.datetime | datetime.date,
) -> bool:
    """Check if the given date is New Year's Day.

    Args:
        date (datetime.datetime | datetime.date): Date to check.

    Returns:
        bool: True if the date is New Year's Day, False otherwise.
    """
    return date.month == 1 and date.day == 1


def is_the_first_three_days_of_new_year(
    date: datetime.datetime | datetime.date,
) -> bool:
    """Check if the given date is within the first three days of the new year.

    Args:
        date (datetime.datetime | datetime.date): Date to check.

    Returns:
        bool: True if the date is within the first three days of the new year, False otherwise.
    """  # noqa: E501
    return date.month == 1 and date.day <= 3


def is_the_end_of_year(
    date: datetime.datetime | datetime.date,
) -> bool:
    """Check if the given date is the end of the year.

    Args:
        date (datetime.datetime | datetime.date): Date to check.

    Returns:
        bool: True if the date is the end of the year, False otherwise.
    """  # noqa: E501
    return date.month == 12 and date.day == 31


def is_between_1231_0103(
    date: datetime.datetime | datetime.date,
) -> bool:
    """Check if the given date is between December 31 and January 3.

    Args:
        date (datetime.datetime | datetime.date): Date to check.

    Returns:
        bool: True if the date is between December 31 and January 3, False otherwise.
    """
    return is_the_end_of_year(date) or is_the_first_three_days_of_new_year(date)


class HolidayDiscriminator:
    """Class to check if a date is a holiday.

    This class allows you to add custom holiday functions and check if a date is a holiday based on those functions.
    It also provides a way to remove holiday functions and get the names of the registered holiday functions.

    Args:
        *funcs (IsHolidayFuncType): Holiday functions to add.
        **kwargs (IsHolidayFuncType): Holiday functions to add.

    Properties:
        names (list[str]): List of names of the registered holiday functions.
        is_holiday_funcs (dict[str, IsHolidayFuncType]): Dictionary of the registered holiday functions.

    Methods:
        __call__(date: datetime.datetime | datetime.date) -> bool: Check if the given date is a holiday.
        add_is_holiday_funcs(*is_holiday_funcs_args: IsHolidayFuncType, **is_holiday_funcs_kwargs: IsHolidayFuncType) -> None: Add custom holiday functions.
        remove_is_holiday_funcs(*names: str) -> None: Remove holiday functions by their names.
    """  # noqa: E501

    def __init__(
        self,
        *funcs: IsHolidayFuncType,
        **kwargs: IsHolidayFuncType,
    ) -> None:
        self._is_holiday_funcs = {func.__name__: func for func in funcs}
        self._is_holiday_funcs.update(kwargs)

    def __call__(self, date: datetime.datetime | datetime.date) -> bool:
        """Check if the given date is a holiday.

        Args:
            date (datetime.datetime | datetime.date): Date to check.

        Returns:
            bool: True if the date is a holiday, False otherwise.
        """
        return any(func(date) for func in self._is_holiday_funcs.values())

    @property
    def names(self) -> list[str]:
        """Get the names of the registered holiday functions.

        Returns:
            list[str]: List of names of the registered holiday functions.
        """
        return list(self._is_holiday_funcs.keys())

    @property
    def is_holiday_funcs(self) -> dict[str, IsHolidayFuncType]:
        """Get the registered holiday functions.

        Returns:
            dict[str, IsHolidayFuncType]: Dictionary of the registered holiday functions.

        Notes:
            The returned dictionary is a shallow copy of the original dictionary.
            Modifications to the returned dictionary will not affect the original dictionary.
        """  # noqa: E501
        return self._is_holiday_funcs.copy()

    def add_is_holiday_funcs(
        self,
        *is_holiday_funcs_args: IsHolidayFuncType,
        allow_overwrite: bool = False,
        **is_holiday_funcs_kwargs: IsHolidayFuncType,
    ) -> None:
        """Add custom holiday functions.

        Args:
            *is_holiday_funcs_args (IsHolidayFuncType): Holiday functions to add.
            allow_overwrite (bool, optional): Allow overwriting existing holiday functions. Defaults to False.
            **is_holiday_funcs_kwargs (IsHolidayFuncType): Holiday functions to add.

        Raises:
            ValueError: If a function with the same name already exists and allow_overwrite is False.
            AttributeError: If the function does not have a __name__ attribute.

        Notes:
            - This method performs an all-or-nothing update: if there is an error, no functions will be added.
        """  # noqa: E501
        # preprocess
        dic: dict[str, IsHolidayFuncType] = {}
        for func in is_holiday_funcs_args:
            try:
                dic[func.__name__] = func
            except AttributeError as e:
                # TODO: more appropriate error handling
                raise e
        dic.update(is_holiday_funcs_kwargs)
        # check duplicates
        for name in dic.keys():
            if not allow_overwrite and name in self._is_holiday_funcs:
                raise ValueError(
                    f"Function with name '{name}' already exists. "
                    "Set allow_overwrite=True to overwrite."
                )
        # add the new functions
        for name, is_holiday_func in dic.items():
            # TODO: more strict type checking
            self._is_holiday_funcs[name] = is_holiday_func

    def remove_is_holiday_funcs(
        self,
        *names: str,
    ) -> None:
        """Remove holiday functions by their names.

        Raises:
            KeyError: If a function with the given name does not exist.

        Notes:
            - This method performs an all-or-nothing update: if there is an error, no functions will be removed.
        """  # noqa: E501
        # check if names are in the dictionary
        for name in names:
            if name not in self._is_holiday_funcs:
                raise KeyError(f"Function with name '{name}' does not exist.")
        # remove the functions
        for name in names:
            self._is_holiday_funcs.pop(name)



import argparse
import contextlib
import time
from datetime import date
from typing import Generator

from pybizday_utils import get_n_next_bizday
from pybizday_utils.holiday_utils import is_saturday_or_sunday


@contextlib.contextmanager
def stopwatch(records: list[float]) -> Generator[None, None, None]:
    """
    Context manager to measure the execution time of a block of code.
    """
    start_time = time.perf_counter()
    try:
        yield
    finally:
        end_time = time.perf_counter()
        records.append(end_time - start_time)


def calc_statistics(records: list[float]) -> dict[str, float]:
    """
    Calculate statistics from a list of execution times.
    """
    # empty case
    if not records:
        return {}
    # calculate statistics
    ave = sum(records) / len(records)
    squared_mean = sum(x ** 2 for x in records) / len(records)
    return {
        "min": min(records),
        "max": max(records),
        "avg": ave,
        "std": (squared_mean - ave ** 2) ** 0.5,
        "count": len(records),
    }


def main() -> None:
    # Parse command line arguments
    parser = argparse.ArgumentParser(description="Check performance of get_n_next_bizday")  # noqa: E501
    parser.add_argument(
        "--n",
        type=int,
        default=100000,
        help="Number of business days to calculate",
    )
    parser.add_argument(
        "--n-trials",
        type=int,
        default=100,
        help="Number of trials to run",
    )
    parser.add_argument(
        "--date",
        type=str,
        default="2023-10-01",
        help="Date to start from (YYYY-MM-DD)",
    )
    args = parser.parse_args()

    # preparation
    n = args.n
    d = date.fromisoformat(args.date)
    n_trials = args.n_trials

    # Execute the function and measure performance
    records = []
    for _ in range(n_trials):
        with stopwatch(records):
            get_n_next_bizday(d, n, is_holiday=is_saturday_or_sunday)

    # Print the results
    stats = calc_statistics(records)
    print(f"Statistics of elapsed time for {n_trials} trials of get_n_next_bizday")  # noqa: E501
    for key, value in stats.items():
        print(f"{key}: {value:.6f}")


if __name__ == "__main__":
    main()



from typing import Generator

import pytest


@pytest.fixture(autouse=True)
def deactivate_global_default_holiday_discriminator(
    request: pytest.FixtureRequest,
) -> Generator[None, None, None]:
    if request.node.get_closest_marker("use_global_default_holiday_discriminator"):  # noqa: E501
        # Skip this fixture if the test is marked with "use_default_holiday_discriminator" # noqa: E501
        yield
    else:
        from pybizday_utils.default_holiday_utils import (
            global_default_holiday_discriminator,
        )  # noqa: I001

        # FIXME: safer way to deactivate the default holiday discriminator.
        #        implement the following codes to avoid use the functions which are tested in test/default_holiday_utils.py  # noqa: E501
        cache = global_default_holiday_discriminator._is_holiday_funcs.copy()
        global_default_holiday_discriminator._is_holiday_funcs = {}
        try:
            yield
        finally:
            global_default_holiday_discriminator._is_holiday_funcs = cache



import datetime
from itertools import dropwhile, filterfalse
from typing import Callable, Generator

from .date_range_utils import date_range
from .default_holiday_utils import global_default_holiday_discriminator
from .holiday_utils import IsHolidayFuncType
from .utils import validate_date_type


def is_bizday(
    date: datetime.date | datetime.datetime,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> bool:
    """Check if the given date is a business day.

    Args:
        date (datetime.date | datetime.datetime): date to check.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is
            a holiday. Defaults to global_default_holiday_discriminator.
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Raises:
        TypeError: If date is not a datetime.date or datetime.datetime.

    Returns:
        bool: True if the date is a business day, False otherwise.
    """
    validate_date_type(date)
    if isinstance(date, datetime.datetime):
        date = datetime_handler(date)
    return not is_holiday(date)


def get_next_bizday(
    date: datetime.date | datetime.datetime,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> datetime.date:
    """Get the next business day after the given date.

    Args:
        date (datetime.date | datetime.datetime): Reference date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is
            a holiday. Defaults to global_default_holiday_discriminator.
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Raises:
        TypeError: If date is not a datetime.date or datetime.datetime.
        ValueError: If no next business day is found.

    Returns:
        datetime.date: Next business day after the given date.
    """
    validate_date_type(date)
    if isinstance(date, datetime.datetime):
        date = datetime_handler(date)
    try:
        return next(dropwhile(is_holiday, date_range(date, include_start=False)))
    except StopIteration as e:
        raise ValueError("No next business day found") from e


def get_prev_bizday(
    date: datetime.date | datetime.datetime,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> datetime.date:
    """Get the previous business day before the given date.

    Args:
        date (datetime.date | datetime.datetime): Reference date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is
            a holiday. Defaults to global_default_holiday_discriminator.
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Raises:
        TypeError: If date is not a datetime.date or datetime.datetime.
        ValueError: If no previous business day is found.

    Returns:
        datetime.date: Previous business day before the given date.
    """
    validate_date_type(date)
    if isinstance(date, datetime.datetime):
        date = datetime_handler(date)
    try:
        return next(
            dropwhile(
                is_holiday,
                date_range(date, include_start=False, step_days=-1),
            )
        )
    except StopIteration as e:
        raise ValueError("No previous business day found") from e


def get_n_next_bizday(
    date: datetime.date | datetime.datetime,
    n: int,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> datetime.date:
    """Get the n-th next business day after the given date.

    Args:
        date (datetime.date | datetime.datetime): Reference date.
        n (int): Number of business days to skip.
            0 means the same date.
            1 means the next business day.
            -1 means the previous business day.
            n > 1 means the n-th next business day.
            n < -1 means the (-n)-th previous business day.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is
            a holiday. Defaults to global_default_holiday_discriminator.
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Raises:
        TypeError: If date is not a datetime.date or datetime.datetime.
        ValueError: If n=0 and the date is a holiday.
        ValueError: If no n-th next business day is found.

    Returns:
        datetime.date: n-th next business day after the given date.

    Notes:
        - If n is negative, it will return the (-n)-th previous business day.
    """
    validate_date_type(date)
    if isinstance(date, datetime.datetime):
        date = datetime_handler(date)
    if n == 0:
        if is_holiday(date):
            raise ValueError(f"n=0 but date={date} is holiday")
        return date
    elif n > 0:
        remaining_days = n
        for d in date_range(date, include_start=False, step_days=1):
            if not is_holiday(d):
                remaining_days -= 1
            if remaining_days == 0:
                return d
        else:
            raise ValueError(f"No {n}-th next business day found")
    else:
        return get_n_prev_bizday(
            date,
            -n,
            is_holiday,
            datetime_handler=datetime_handler,
        )


def get_n_prev_bizday(
    date: datetime.date | datetime.datetime,
    n: int,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> datetime.date:
    """Get the n-th previous business day before the given date."

    Args:
        date (datetime.date | datetime.datetime): Reference date.
        n (int): Number of business days to skip.
            0 means the same date.
            1 means the previous business day.
            -1 means the next business day.
            n > 1 means the n-th previous business day.
            n < -1 means the (-n)-th next business day.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is
            a holiday. Defaults to global_default_holiday_discriminator.
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Raises:
        TypeError: If date is not a datetime.date or datetime.datetime.
        ValueError: If n=0 and the date is a holiday.
        ValueError: If no n-th previous business day is found.

    Returns:
        datetime.date: n-th previous business day before the given date.

    Notes:
        - If n is negative, it will return the (-n)-th next business day.
    """
    validate_date_type(date)
    if isinstance(date, datetime.datetime):
        date = datetime_handler(date)
    if n == 0:
        if is_holiday(date):
            raise ValueError(f"n=0 but date={date} is holiday")
        return date
    elif n > 0:
        remaining_days = n
        for d in date_range(date, include_start=False, step_days=-1):
            if not is_holiday(d):
                remaining_days -= 1
            if remaining_days == 0:
                return d
        else:
            raise ValueError(f"No {n}-th previous business day found")
    else:
        return get_n_next_bizday(
            date,
            -n,
            is_holiday,
            datetime_handler=datetime_handler,
        )


def bizday_range(
    start: datetime.date | datetime.datetime,
    end: datetime.date | datetime.datetime,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    include_start: bool = True,
    include_end: bool = True,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> Generator[datetime.date, None, None]:
    """Generate a range of business days between two dates."

    Args:
        start (datetime.date | datetime.datetime): Start date.
        end (datetime.date | datetime.datetime): End date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is
            a holiday. Defaults to global_default_holiday_discriminator.
        include_start (bool, optional): Include the start date in the range.
            Defaults to True.
        include_end (bool, optional): Include the end date in the range.
            Defaults to True.
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Raises:
        TypeError: If start or end is not a datetime.date or datetime.datetime.

    Yields:
        Generator[datetime.date, None, None]: Business days between start and end
            dates.

    Notes:
        - if include_start is True and start is not a holiday, the start date
          will be included in the range.
        - if include_end is True and end is not a holiday, the end date will be
          included in the range.
    """
    validate_date_type(start)
    validate_date_type(end)
    if isinstance(start, datetime.datetime):
        start = datetime_handler(start)
    if isinstance(end, datetime.datetime):
        end = datetime_handler(end)
    yield from filterfalse(
        is_holiday,
        date_range(
            start,
            end,
            include_start=include_start,
            include_end=include_end,
            step_days=1 if start <= end else -1,
        ),
    )


def count_bizdays(
    start: datetime.date | datetime.datetime,
    end: datetime.date | datetime.datetime,
    is_holiday: IsHolidayFuncType = global_default_holiday_discriminator,
    *,
    include_start: bool = True,
    include_end: bool = True,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> int:
    """Count the number of business days between two dates.

    Args:
        start (datetime.date | datetime.datetime): Start date.
        end (datetime.date | datetime.datetime): End date.
        is_holiday (IsHolidayFuncType, optional): function to check if a date is
            a holiday. Defaults to global_default_holiday_discriminator.
        include_start (bool, optional): Include the start date in the count.
            Defaults to True.
        include_end (bool, optional): Include the end date in the count.
            Defaults to True.
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Raises:
        TypeError: If date is not a datetime.date or datetime.datetime.

    Returns:
        int: Number of business days between start and end dates.

    Notes:
        - if start > end, the count will be negative.
        - if include_start is True and start is not a holiday, the start date
          will be included in the count.
        - if include_end is True and end is not a holiday, the end date will be
          included in the count.
    """
    validate_date_type(start)
    validate_date_type(end)
    if isinstance(start, datetime.datetime):
        start = datetime_handler(start)
    if isinstance(end, datetime.datetime):
        end = datetime_handler(end)
    if start > end:
        return -count_bizdays(
            end,
            start,
            is_holiday,
            include_end=include_start,
            include_start=include_end,
            datetime_handler=datetime_handler,
        )
    bdrange = bizday_range(
        start,
        end,
        is_holiday,
        include_start=include_start,
        include_end=include_end,
        datetime_handler=datetime_handler,
    )
    return sum(1 for _ in bdrange)



import datetime
from typing import Any

import pytest

from pybizday_utils.utils import validate_date_type


@pytest.mark.parametrize(
    "date",
    [
        datetime.date(2025, 1, 1),
        datetime.datetime(2025, 1, 1),
        datetime.datetime(2025, 1, 1, 12, 0),
    ],
)
def test_validate_date_type_with_valid_types(
    date: datetime.date | datetime.datetime,
) -> None:
    """Test validate_date_type with valid date types."""
    assert validate_date_type(date) is True


@pytest.mark.parametrize(
    "invalid_date",
    [
        "2025-01-01",  # string
        20250101,  # integer
        1.5,  # float
        None,  # None
        [],  # list
        {},  # dict
        (2025, 1, 1),  # tuple
    ],
)
def test_validate_date_type_with_invalid_types(invalid_date: Any) -> None:
    """Test validate_date_type with invalid date types."""
    with pytest.raises(TypeError) as exc_info:
        validate_date_type(invalid_date)
    assert "must be a datetime.date or datetime.datetime object" in str(exc_info.value)  # noqa: E501
    assert str(type(invalid_date)) in str(exc_info.value)



from . import (
    default_holiday_utils,
    holiday_utils,
)
from .basic import (
    bizday_range,
    count_bizdays,
    get_n_next_bizday,
    get_n_prev_bizday,
    get_next_bizday,
    get_prev_bizday,
    is_bizday,
)
from .month import (
    add_months,
    add_years,
    add_years_months,
    get_biz_end_of_month,
    get_biz_start_of_month,
    is_biz_end_of_month,
    is_biz_start_of_month,
)

try:
    from ._version import __version__  # noqa
except ImportError as e:
    raise ImportError(
        "Failed to import the version information for pybizday_utils. "
        "This may be due to a deployment issue. "
        "Please report this to the developer: https://github.com/hmasdev/pybizday_utils/issues/new"  # noqa: E501
    ) from e

__all__ = [
    "add_months",
    "add_years",
    "add_years_months",
    "get_biz_end_of_month",
    "get_biz_start_of_month",
    "is_biz_end_of_month",
    "is_biz_start_of_month",
    "bizday_range",
    "count_bizdays",
    "get_n_next_bizday",
    "get_n_prev_bizday",
    "get_next_bizday",
    "get_prev_bizday",
    "is_bizday",
    "default_holiday_utils",
    "holiday_utils",
]



from datetime import date, datetime

import pytest

from pybizday_utils.holiday_utils import (
    HolidayDiscriminator,
    is_between_1231_0103,
    is_new_year_day,
    is_saturday_or_sunday,
    is_the_end_of_year,
    is_the_first_three_days_of_new_year,
)


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, expected",
    [
        (date(2025, 3, 21), False),
        (date(2025, 3, 22), True),
        (date(2025, 3, 23), True),
        (date(2025, 3, 24), False),
        (datetime(2025, 3, 21), False),
        (datetime(2025, 3, 22), True),
        (datetime(2025, 3, 23), True),
        (datetime(2025, 3, 24), False),
    ],
)
def test_is_saturday_or_sunday(
    d: datetime | date,
    expected: bool,
) -> None:
    assert is_saturday_or_sunday(d) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, expected",
    [
        (date(2025, 1, 1), True),
        (date(2025, 1, 2), False),
        (datetime(2024, 1, 1), True),
        (datetime(2024, 1, 2), False),
    ],
)
def test_is_new_year_day(
    d: datetime | date,
    expected: bool,
) -> None:
    assert is_new_year_day(d) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, expected",
    [
        (date(2024, 12, 31), False),
        (date(2025, 1, 1), True),
        (date(2025, 1, 2), True),
        (date(2025, 1, 3), True),
        (date(2025, 1, 4), False),
        (datetime(2024, 12, 31), False),
        (datetime(2024, 1, 1), True),
        (datetime(2024, 1, 2), True),
        (datetime(2024, 1, 3), True),
        (datetime(2024, 1, 4), False),
    ],
)
def test_is_the_first_three_days_of_new_year(
    d: datetime | date,
    expected: bool,
) -> None:
    assert is_the_first_three_days_of_new_year(d) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, expected",
    [
        (date(2024, 12, 30), False),
        (date(2024, 12, 31), True),
        (date(2025, 1, 1), False),
        (date(2025, 1, 2), False),
        (datetime(2024, 12, 30), False),
        (datetime(2024, 12, 31), True),
        (datetime(2025, 1, 1), False),
        (datetime(2025, 1, 2), False),
    ],
)
def test_is_the_end_of_year(
    d: datetime | date,
    expected: bool,
) -> None:
    assert is_the_end_of_year(d) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, expected",
    [
        (date(2024, 12, 30), False),
        (date(2024, 12, 31), True),
        (date(2025, 1, 1), True),
        (date(2025, 1, 2), True),
        (date(2025, 1, 3), True),
        (date(2025, 1, 4), False),
        (datetime(2024, 12, 30), False),
        (datetime(2024, 12, 31), True),
        (datetime(2025, 1, 1), True),
        (datetime(2025, 1, 2), True),
        (datetime(2025, 1, 3), True),
        (datetime(2025, 1, 4), False),
    ],
)
def test_is_between_1231_0103(
    d: datetime | date,
    expected: bool,
) -> None:
    assert is_between_1231_0103(d) == expected


@pytest.mark.positive
def test_holiday_discriminator_with_arg() -> None:
    discriminator = HolidayDiscriminator(is_saturday_or_sunday)
    assert discriminator(date(2025, 3, 21)) is False
    assert discriminator(date(2025, 3, 22)) is True
    assert discriminator(date(2025, 3, 23)) is True
    assert discriminator(date(2025, 3, 24)) is False
    assert discriminator.names == [is_saturday_or_sunday.__name__]


@pytest.mark.positive
def test_holiday_discriminator_with_kwarg() -> None:
    discriminator = HolidayDiscriminator(is_sos=is_saturday_or_sunday)
    assert discriminator(date(2025, 3, 21)) is False
    assert discriminator(date(2025, 3, 22)) is True
    assert discriminator(date(2025, 3, 23)) is True
    assert discriminator(date(2025, 3, 24)) is False
    assert discriminator.names == ["is_sos"]


@pytest.mark.positive
def test_holiday_discriminator_with_args_kwargs() -> None:
    discriminator = HolidayDiscriminator(
        is_saturday_or_sunday,
        is_new_year_day,
        is_eoy=is_the_end_of_year,
        is_20250325=lambda d: d == date(2025, 3, 25),
    )
    assert discriminator(date(2025, 3, 21)) is False
    assert discriminator(date(2025, 3, 22)) is True
    assert discriminator(date(2025, 3, 23)) is True
    assert discriminator(date(2025, 3, 24)) is False
    assert discriminator(date(2025, 1, 1)) is True
    assert discriminator(date(2025, 1, 2)) is False
    assert discriminator(date(2024, 12, 30)) is False
    assert discriminator(date(2024, 12, 31)) is True
    assert discriminator(date(2025, 3, 25)) is True
    assert discriminator(date(2025, 3, 26)) is False
    assert discriminator.names == [
        is_saturday_or_sunday.__name__,
        is_new_year_day.__name__,
        "is_eoy",
        "is_20250325",
    ]


@pytest.mark.positive
def test_holiday_discriminator_kwargs_priority() -> None:
    def dummy(date: date) -> bool:
        raise NotImplementedError()

    # use dummy function when dummy is not overwritten by kwargs
    discriminator = HolidayDiscriminator(dummy)
    with pytest.raises(NotImplementedError):
        discriminator(date(2025, 1, 1))
    assert discriminator.names == ["dummy"]

    # use is_new_year_day when dummy is overwritten by kwargs
    discriminator = HolidayDiscriminator(dummy, dummy=is_new_year_day)  # noqa: E501
    assert discriminator(date(2025, 1, 1)) is True
    assert discriminator(date(2025, 1, 2)) is False
    assert discriminator.names == ["dummy"]


@pytest.mark.positive
def test_holiday_discriminator_properties() -> None:
    discriminator = HolidayDiscriminator(is_new_year_day)
    assert discriminator.names == list(discriminator._is_holiday_funcs.keys())
    assert discriminator.is_holiday_funcs == discriminator._is_holiday_funcs

    discriminator._is_holiday_funcs["func"] = is_saturday_or_sunday
    assert discriminator.names == list(discriminator._is_holiday_funcs.keys())
    assert discriminator.is_holiday_funcs == discriminator._is_holiday_funcs

    discriminator._is_holiday_funcs["func"] = is_between_1231_0103
    assert discriminator.names == list(discriminator._is_holiday_funcs.keys())
    assert discriminator.is_holiday_funcs == discriminator._is_holiday_funcs

    discriminator._is_holiday_funcs["func2"] = is_the_end_of_year
    assert discriminator.names == list(discriminator._is_holiday_funcs.keys())
    assert discriminator.is_holiday_funcs == discriminator._is_holiday_funcs

    discriminator._is_holiday_funcs.pop("func")
    assert discriminator.names == list(discriminator._is_holiday_funcs.keys())
    assert discriminator.is_holiday_funcs == discriminator._is_holiday_funcs


@pytest.mark.positive
def test_holiday_discriminator_add_is_holiday_funcs_as_args() -> None:
    discriminator = HolidayDiscriminator()
    assert discriminator(date(2025, 1, 1)) is False
    assert discriminator.names == []
    cache = discriminator._is_holiday_funcs.copy()

    discriminator.add_is_holiday_funcs(is_new_year_day)
    assert discriminator(date(2025, 1, 1)) is True
    assert discriminator.names == [is_new_year_day.__name__]

    # Just in case
    assert discriminator._is_holiday_funcs != cache


@pytest.mark.positive
def test_holiday_discriminator_add_is_holiday_funcs_as_kwargs() -> None:
    discriminator = HolidayDiscriminator()
    assert discriminator(date(2025, 1, 1)) is False
    assert discriminator.names == []
    cache = discriminator._is_holiday_funcs.copy()

    discriminator.add_is_holiday_funcs(inyd=is_new_year_day)
    assert discriminator(date(2025, 1, 1)) is True
    assert discriminator.names == ["inyd"]

    # Just in case
    assert discriminator._is_holiday_funcs != cache


@pytest.mark.positive
def test_holiday_discriminator_add_is_holiday_funcs_with_allow_overwrite() -> None:
    discriminator = HolidayDiscriminator(func=is_new_year_day)
    assert discriminator(date(2025, 1, 1)) is True
    assert discriminator.names == ["func"]
    cache = discriminator._is_holiday_funcs.copy()

    discriminator.add_is_holiday_funcs(func=is_saturday_or_sunday, allow_overwrite=True)  # noqa
    assert discriminator(date(2025, 1, 1)) is False
    assert discriminator.names == ["func"]

    # Just in case
    assert discriminator._is_holiday_funcs != cache


@pytest.mark.negative
def test_holiday_discriminator_add_is_holiday_funcs_should_prohibit_overwrite_in_default() -> (  # noqa: E501
    None
):
    discriminator = HolidayDiscriminator(func=is_new_year_day)
    assert discriminator(date(2025, 1, 1)) is True
    assert discriminator.names == ["func"]
    cache = discriminator._is_holiday_funcs.copy()

    with pytest.raises(ValueError):
        discriminator.add_is_holiday_funcs(
            is_the_end_of_year,
            funb=is_the_end_of_year,
            func=is_saturday_or_sunday,
            fund=is_new_year_day,
        )
    assert discriminator(date(2025, 1, 1)) is True

    # NOTE: Ensure all or nothing is updated. # noqa: E501
    assert discriminator.names == ["func"]

    # Just in case
    # NOTE: Ensure all or nothing is updated.
    assert discriminator._is_holiday_funcs == cache


@pytest.mark.negative
def test_holiday_discriminator_add_is_holiday_funcs_as_args_without_name_attr() -> None:  # noqa: E501
    class CallableCls:
        def __call__(self, d: date) -> bool:
            return d == d

    callable_instance = CallableCls()
    assert not hasattr(callable_instance, "__name__")

    discriminator = HolidayDiscriminator()
    cache = discriminator._is_holiday_funcs.copy()

    with pytest.raises(AttributeError):
        discriminator.add_is_holiday_funcs(
            is_new_year_day,
            callable_instance,
            is_saturday_or_sunday,
            func=is_the_end_of_year,
        )

    # NOTE: Ensure all or nothing is updated.
    assert discriminator.names == []

    # Just in case
    # NOTE: Ensure all or nothing is updated.
    assert discriminator._is_holiday_funcs == cache


@pytest.mark.positive
def test_holiday_discriminator_remove() -> None:
    discriminator = HolidayDiscriminator(is_new_year_day)
    assert discriminator(date(2025, 1, 1)) is True
    assert discriminator.names == [is_new_year_day.__name__]

    discriminator.remove_is_holiday_funcs(is_new_year_day.__name__)
    assert discriminator(date(2025, 1, 1)) is False
    assert discriminator.names == []


@pytest.mark.negative
def test_holiday_discriminator_remove_not_exist() -> None:
    discriminator = HolidayDiscriminator(is_new_year_day)
    assert discriminator.names == [is_new_year_day.__name__]
    with pytest.raises(KeyError):
        discriminator.remove_is_holiday_funcs(
            is_new_year_day.__name__,
            is_saturday_or_sunday.__name__,
        )

    # NOTE: Ensure all or nothing is updated.
    assert discriminator.names == [is_new_year_day.__name__]



from datetime import date, datetime, timedelta
from typing import Callable

import pytest
from pytest_mock import MockerFixture

import pybizday_utils.month
from pybizday_utils.month import (
    add_months,
    add_years,
    add_years_months,
    get_biz_end_of_month,
    get_biz_start_of_month,
    is_biz_end_of_month,
    is_biz_start_of_month,
)


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,expected",
    [
        # Case: everyday is not holiday
        (date(2023, 1, 31), lambda _: False, True),
        (date(2023, 2, 28), lambda _: False, True),
        (date(2023, 3, 31), lambda _: False, True),
        (date(2023, 4, 30), lambda _: False, True),
        (date(2023, 5, 31), lambda _: False, True),
        (date(2023, 6, 30), lambda _: False, True),
        (date(2023, 7, 31), lambda _: False, True),
        (date(2023, 8, 31), lambda _: False, True),
        (date(2023, 9, 30), lambda _: False, True),
        (date(2023, 10, 31), lambda _: False, True),
        (date(2023, 11, 30), lambda _: False, True),
        (date(2023, 12, 31), lambda _: False, True),
        (date(2023, 1, 30), lambda _: False, False),
        (date(2023, 2, 27), lambda _: False, False),
        (date(2023, 3, 30), lambda _: False, False),
        (date(2023, 4, 29), lambda _: False, False),
        (date(2023, 5, 30), lambda _: False, False),
        (date(2023, 6, 29), lambda _: False, False),
        (date(2023, 7, 30), lambda _: False, False),
        (date(2023, 8, 30), lambda _: False, False),
        (date(2023, 9, 29), lambda _: False, False),
        (date(2023, 10, 30), lambda _: False, False),
        (date(2023, 11, 29), lambda _: False, False),
        (date(2023, 12, 30), lambda _: False, False),
        # Case: specify is_holiday function
        (date(2023, 1, 31), lambda d: d == date(2023, 1, 31), False),
        (date(2023, 1, 30), lambda d: d == date(2023, 1, 31), True),
        (date(2023, 1, 29), lambda d: d == date(2023, 1, 31), False),
        (date(2023, 2, 28), lambda d: d == date(2023, 2, 28), False),
        (date(2023, 2, 27), lambda d: d == date(2023, 2, 28), True),
        (date(2023, 2, 26), lambda d: d == date(2023, 2, 28), False),
        (date(2023, 3, 31), lambda d: d == date(2023, 3, 31), False),
        (date(2023, 3, 30), lambda d: d == date(2023, 3, 31), True),
        (date(2023, 3, 29), lambda d: d == date(2023, 3, 31), False),
        (date(2023, 4, 30), lambda d: d == date(2023, 4, 30), False),
        (date(2023, 4, 29), lambda d: d == date(2023, 4, 30), True),
        (date(2023, 4, 28), lambda d: d == date(2023, 4, 30), False),
        (date(2023, 5, 31), lambda d: d == date(2023, 5, 31), False),
        (date(2023, 5, 30), lambda d: d == date(2023, 5, 31), True),
        (date(2023, 5, 29), lambda d: d == date(2023, 5, 31), False),
        (date(2023, 6, 30), lambda d: d == date(2023, 6, 30), False),
        (date(2023, 6, 29), lambda d: d == date(2023, 6, 30), True),
        (date(2023, 6, 28), lambda d: d == date(2023, 6, 30), False),
        (date(2023, 7, 31), lambda d: d == date(2023, 7, 31), False),
        (date(2023, 7, 30), lambda d: d == date(2023, 7, 31), True),
        (date(2023, 7, 29), lambda d: d == date(2023, 7, 31), False),
        (date(2023, 8, 31), lambda d: d == date(2023, 8, 31), False),
        (date(2023, 8, 30), lambda d: d == date(2023, 8, 31), True),
        (date(2023, 8, 29), lambda d: d == date(2023, 8, 31), False),
        (date(2023, 9, 30), lambda d: d == date(2023, 9, 30), False),
        (date(2023, 9, 29), lambda d: d == date(2023, 9, 30), True),
        (date(2023, 9, 28), lambda d: d == date(2023, 9, 30), False),
        (date(2023, 10, 31), lambda d: d == date(2023, 10, 31), False),
        (date(2023, 10, 30), lambda d: d == date(2023, 10, 31), True),
        (date(2023, 10, 29), lambda d: d == date(2023, 10, 31), False),
        (date(2023, 11, 30), lambda d: d == date(2023, 11, 30), False),
        (date(2023, 11, 29), lambda d: d == date(2023, 11, 30), True),
        (date(2023, 11, 28), lambda d: d == date(2023, 11, 30), False),
        (date(2023, 12, 31), lambda d: d == date(2023, 12, 31), False),
        (date(2023, 12, 30), lambda d: d == date(2023, 12, 31), True),
        (date(2023, 12, 29), lambda d: d == date(2023, 12, 31), False),
        # Case: actual. Saturday and Sunday are holidays
        (date(2024, 11, 30), lambda d: d.weekday() in {5, 6}, False),
        (date(2024, 11, 29), lambda d: d.weekday() in {5, 6}, True),
        (date(2024, 11, 28), lambda d: d.weekday() in {5, 6}, False),
        (date(2025, 1, 31), lambda d: d.weekday() in {5, 6}, True),
        (date(2025, 1, 30), lambda d: d.weekday() in {5, 6}, False),
    ],
)
def test_is_biz_end_of_month(
    d: date,
    is_holiday: Callable[[date], bool],
    expected: bool,
) -> None:
    assert is_biz_end_of_month(d, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,expected",
    [
        # Case: everyday is not holiday
        (date(2023, 1, 1), lambda _: False, True),
        (date(2023, 2, 1), lambda _: False, True),
        (date(2023, 3, 1), lambda _: False, True),
        (date(2023, 4, 1), lambda _: False, True),
        (date(2023, 5, 1), lambda _: False, True),
        (date(2023, 6, 1), lambda _: False, True),
        (date(2023, 7, 1), lambda _: False, True),
        (date(2023, 8, 1), lambda _: False, True),
        (date(2023, 9, 1), lambda _: False, True),
        (date(2023, 10, 1), lambda _: False, True),
        (date(2023, 11, 1), lambda _: False, True),
        (date(2023, 12, 1), lambda _: False, True),
        (date(2023, 1, 2), lambda _: False, False),
        (date(2023, 2, 2), lambda _: False, False),
        (date(2023, 3, 2), lambda _: False, False),
        (date(2023, 4, 2), lambda _: False, False),
        (date(2023, 5, 2), lambda _: False, False),
        (date(2023, 6, 2), lambda _: False, False),
        (date(2023, 7, 2), lambda _: False, False),
        (date(2023, 8, 2), lambda _: False, False),
        (date(2023, 9, 2), lambda _: False, False),
        (date(2023, 10, 2), lambda _: False, False),
        (date(2023, 11, 2), lambda _: False, False),
        (date(2023, 12, 2), lambda _: False, False),
        # Case: specify is_holiday function
        (date(2023, 1, 1), lambda d: d == date(2023, 1, 1), False),
        (date(2023, 1, 2), lambda d: d == date(2023, 1, 1), True),
        (date(2023, 1, 3), lambda d: d == date(2023, 1, 1), False),
        (date(2023, 2, 1), lambda d: d == date(2023, 2, 1), False),
        (date(2023, 2, 2), lambda d: d == date(2023, 2, 1), True),
        (date(2023, 2, 3), lambda d: d == date(2023, 2, 1), False),
        (date(2023, 3, 1), lambda d: d == date(2023, 3, 1), False),
        (date(2023, 3, 2), lambda d: d == date(2023, 3, 1), True),
        (date(2023, 3, 3), lambda d: d == date(2023, 3, 1), False),
        (date(2023, 4, 1), lambda d: d == date(2023, 4, 1), False),
        (date(2023, 4, 2), lambda d: d == date(2023, 4, 1), True),
        (date(2023, 4, 3), lambda d: d == date(2023, 4, 1), False),
        (date(2023, 5, 1), lambda d: d == date(2023, 5, 1), False),
        (date(2023, 5, 2), lambda d: d == date(2023, 5, 1), True),
        (date(2023, 5, 3), lambda d: d == date(2023, 5, 1), False),
        (date(2023, 6, 1), lambda d: d == date(2023, 6, 1), False),
        (date(2023, 6, 2), lambda d: d == date(2023, 6, 1), True),
        (date(2023, 6, 3), lambda d: d == date(2023, 6, 1), False),
        (date(2023, 7, 1), lambda d: d == date(2023, 7, 1), False),
        (date(2023, 7, 2), lambda d: d == date(2023, 7, 1), True),
        (date(2023, 7, 3), lambda d: d == date(2023, 7, 1), False),
        (date(2023, 8, 1), lambda d: d == date(2023, 8, 1), False),
        (date(2023, 8, 2), lambda d: d == date(2023, 8, 1), True),
        (date(2023, 8, 3), lambda d: d == date(2023, 8, 1), False),
        (date(2023, 9, 1), lambda d: d == date(2023, 9, 1), False),
        (date(2023, 9, 2), lambda d: d == date(2023, 9, 1), True),
        (date(2023, 9, 3), lambda d: d == date(2023, 9, 1), False),
        (date(2023, 10, 1), lambda d: d == date(2023, 10, 1), False),
        (date(2023, 10, 2), lambda d: d == date(2023, 10, 1), True),
        (date(2023, 10, 3), lambda d: d == date(2023, 10, 1), False),
        (date(2023, 11, 1), lambda d: d == date(2023, 11, 1), False),
        (date(2023, 11, 2), lambda d: d == date(2023, 11, 1), True),
        (date(2023, 11, 3), lambda d: d == date(2023, 11, 1), False),
        (date(2023, 12, 1), lambda d: d == date(2023, 12, 1), False),
        (date(2023, 12, 2), lambda d: d == date(2023, 12, 1), True),
        (date(2023, 12, 3), lambda d: d == date(2023, 12, 1), False),
        # Case: actual. Saturday and Sunday are holidays
        (date(2024, 12, 1), lambda d: d.weekday() in {5, 6}, False),
        (date(2024, 12, 2), lambda d: d.weekday() in {5, 6}, True),
        (date(2024, 12, 3), lambda d: d.weekday() in {5, 6}, False),
        (date(2024, 11, 1), lambda d: d.weekday() in {5, 6}, True),
        (date(2024, 11, 2), lambda d: d.weekday() in {5, 6}, False),
    ],
)
def test_is_biz_start_of_month(
    d: date,
    is_holiday: Callable[[date], bool],
    expected: bool,
) -> None:
    assert is_biz_start_of_month(d, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,expected",
    [
        # Case: everyday is not holiday
        (date(2023, 1, 1), lambda _: False, date(2023, 1, 1)),
        (date(2023, 1, 31), lambda _: False, date(2023, 1, 1)),
        (date(2023, 2, 1), lambda _: False, date(2023, 2, 1)),
        (date(2023, 2, 28), lambda _: False, date(2023, 2, 1)),
        (date(2023, 3, 1), lambda _: False, date(2023, 3, 1)),
        (date(2023, 3, 31), lambda _: False, date(2023, 3, 1)),
        (date(2023, 4, 1), lambda _: False, date(2023, 4, 1)),
        (date(2023, 4, 30), lambda _: False, date(2023, 4, 1)),
        (date(2023, 5, 1), lambda _: False, date(2023, 5, 1)),
        (date(2023, 5, 31), lambda _: False, date(2023, 5, 1)),
        (date(2023, 6, 1), lambda _: False, date(2023, 6, 1)),
        (date(2023, 6, 30), lambda _: False, date(2023, 6, 1)),
        (date(2023, 7, 1), lambda _: False, date(2023, 7, 1)),
        (date(2023, 7, 31), lambda _: False, date(2023, 7, 1)),
        (date(2023, 8, 1), lambda _: False, date(2023, 8, 1)),
        (date(2023, 8, 31), lambda _: False, date(2023, 8, 1)),
        (date(2023, 9, 1), lambda _: False, date(2023, 9, 1)),
        (date(2023, 9, 30), lambda _: False, date(2023, 9, 1)),
        (date(2023, 10, 1), lambda _: False, date(2023, 10, 1)),
        (date(2023, 10, 31), lambda _: False, date(2023, 10, 1)),
        (date(2023, 11, 1), lambda _: False, date(2023, 11, 1)),
        (date(2023, 11, 30), lambda _: False, date(2023, 11, 1)),
        (date(2023, 12, 1), lambda _: False, date(2023, 12, 1)),
        (date(2023, 12, 31), lambda _: False, date(2023, 12, 1)),
        # Case: specify is_holiday function
        (date(2023, 1, 1), lambda d: d == date(2023, 1, 1), date(2023, 1, 2)),
        (date(2023, 1, 2), lambda d: d == date(2023, 1, 1), date(2023, 1, 2)),
        (date(2023, 1, 31), lambda d: d == date(2023, 1, 1), date(2023, 1, 2)),
        (date(2023, 2, 1), lambda d: d == date(2023, 2, 1), date(2023, 2, 2)),
        (date(2023, 2, 2), lambda d: d == date(2023, 2, 1), date(2023, 2, 2)),
        (date(2023, 2, 28), lambda d: d == date(2023, 2, 1), date(2023, 2, 2)),
        (date(2023, 3, 1), lambda d: d == date(2023, 3, 1), date(2023, 3, 2)),
        (date(2023, 3, 2), lambda d: d == date(2023, 3, 1), date(2023, 3, 2)),
        (date(2023, 3, 31), lambda d: d == date(2023, 3, 1), date(2023, 3, 2)),
        (date(2023, 4, 1), lambda d: d == date(2023, 4, 1), date(2023, 4, 2)),
        (date(2023, 4, 2), lambda d: d == date(2023, 4, 1), date(2023, 4, 2)),
        (date(2023, 4, 30), lambda d: d == date(2023, 4, 1), date(2023, 4, 2)),
        (date(2023, 5, 1), lambda d: d == date(2023, 5, 1), date(2023, 5, 2)),
        (date(2023, 5, 2), lambda d: d == date(2023, 5, 1), date(2023, 5, 2)),
        (date(2023, 5, 31), lambda d: d == date(2023, 5, 1), date(2023, 5, 2)),
        (date(2023, 6, 1), lambda d: d == date(2023, 6, 1), date(2023, 6, 2)),
        (date(2023, 6, 2), lambda d: d == date(2023, 6, 1), date(2023, 6, 2)),
        (date(2023, 6, 30), lambda d: d == date(2023, 6, 1), date(2023, 6, 2)),
        (date(2023, 7, 1), lambda d: d == date(2023, 7, 1), date(2023, 7, 2)),
        (date(2023, 7, 2), lambda d: d == date(2023, 7, 1), date(2023, 7, 2)),
        (date(2023, 7, 31), lambda d: d == date(2023, 7, 1), date(2023, 7, 2)),
        (date(2023, 8, 1), lambda d: d == date(2023, 8, 1), date(2023, 8, 2)),
        (date(2023, 8, 2), lambda d: d == date(2023, 8, 1), date(2023, 8, 2)),
        (date(2023, 8, 31), lambda d: d == date(2023, 8, 1), date(2023, 8, 2)),
        (date(2023, 9, 1), lambda d: d == date(2023, 9, 1), date(2023, 9, 2)),
        (date(2023, 9, 2), lambda d: d == date(2023, 9, 1), date(2023, 9, 2)),
        (date(2023, 9, 30), lambda d: d == date(2023, 9, 1), date(2023, 9, 2)),
        (date(2023, 10, 1), lambda d: d == date(2023, 10, 1), date(2023, 10, 2)),
        (date(2023, 10, 2), lambda d: d == date(2023, 10, 1), date(2023, 10, 2)),
        (date(2023, 10, 31), lambda d: d == date(2023, 10, 1), date(2023, 10, 2)),
        (date(2023, 11, 1), lambda d: d == date(2023, 11, 1), date(2023, 11, 2)),
        (date(2023, 11, 2), lambda d: d == date(2023, 11, 1), date(2023, 11, 2)),
        (date(2023, 11, 30), lambda d: d == date(2023, 11, 1), date(2023, 11, 2)),
        (date(2023, 12, 1), lambda d: d == date(2023, 12, 1), date(2023, 12, 2)),
        (date(2023, 12, 2), lambda d: d == date(2023, 12, 1), date(2023, 12, 2)),
        (date(2023, 12, 31), lambda d: d == date(2023, 12, 1), date(2023, 12, 2)),
    ],
)
def test_get_biz_start_of_month(
    d: date,
    is_holiday: Callable[[date], bool],
    expected: date,
) -> None:
    assert get_biz_start_of_month(d, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,expected",
    [
        # Case: everyday is not holiday
        (date(2023, 1, 1), lambda _: False, date(2023, 1, 31)),
        (date(2023, 1, 31), lambda _: False, date(2023, 1, 31)),
        (date(2023, 2, 1), lambda _: False, date(2023, 2, 28)),
        (date(2023, 2, 28), lambda _: False, date(2023, 2, 28)),
        (date(2023, 3, 1), lambda _: False, date(2023, 3, 31)),
        (date(2023, 3, 31), lambda _: False, date(2023, 3, 31)),
        (date(2023, 4, 1), lambda _: False, date(2023, 4, 30)),
        (date(2023, 4, 30), lambda _: False, date(2023, 4, 30)),
        (date(2023, 5, 1), lambda _: False, date(2023, 5, 31)),
        (date(2023, 5, 31), lambda _: False, date(2023, 5, 31)),
        (date(2023, 6, 1), lambda _: False, date(2023, 6, 30)),
        (date(2023, 6, 30), lambda _: False, date(2023, 6, 30)),
        (date(2023, 7, 1), lambda _: False, date(2023, 7, 31)),
        (date(2023, 7, 31), lambda _: False, date(2023, 7, 31)),
        (date(2023, 8, 1), lambda _: False, date(2023, 8, 31)),
        (date(2023, 8, 31), lambda _: False, date(2023, 8, 31)),
        (date(2023, 9, 1), lambda _: False, date(2023, 9, 30)),
        (date(2023, 9, 30), lambda _: False, date(2023, 9, 30)),
        (date(2023, 10, 1), lambda _: False, date(2023, 10, 31)),
        (date(2023, 10, 31), lambda _: False, date(2023, 10, 31)),
        (date(2023, 11, 1), lambda _: False, date(2023, 11, 30)),
        (date(2023, 11, 30), lambda _: False, date(2023, 11, 30)),
        (date(2023, 12, 1), lambda _: False, date(2023, 12, 31)),
        (date(2023, 12, 31), lambda _: False, date(2023, 12, 31)),
        (date(2024, 2, 1), lambda _: False, date(2024, 2, 29)),
        (date(2024, 2, 29), lambda _: False, date(2024, 2, 29)),
        # Case: specify is_holiday function
        (date(2023, 1, 1), lambda d: d == date(2023, 1, 31), date(2023, 1, 30)),
        (date(2023, 1, 30), lambda d: d == date(2023, 1, 31), date(2023, 1, 30)),
        (date(2023, 1, 31), lambda d: d == date(2023, 1, 31), date(2023, 1, 30)),
        (date(2023, 2, 1), lambda d: d == date(2023, 2, 28), date(2023, 2, 27)),
        (date(2023, 2, 27), lambda d: d == date(2023, 2, 28), date(2023, 2, 27)),
        (date(2023, 2, 28), lambda d: d == date(2023, 2, 28), date(2023, 2, 27)),
        (date(2023, 3, 1), lambda d: d == date(2023, 3, 31), date(2023, 3, 30)),
        (date(2023, 3, 30), lambda d: d == date(2023, 3, 31), date(2023, 3, 30)),
        (date(2023, 3, 31), lambda d: d == date(2023, 3, 31), date(2023, 3, 30)),
        (date(2023, 4, 1), lambda d: d == date(2023, 4, 30), date(2023, 4, 29)),
        (date(2023, 4, 29), lambda d: d == date(2023, 4, 30), date(2023, 4, 29)),
        (date(2023, 4, 30), lambda d: d == date(2023, 4, 30), date(2023, 4, 29)),
        (date(2023, 5, 1), lambda d: d == date(2023, 5, 31), date(2023, 5, 30)),
        (date(2023, 5, 30), lambda d: d == date(2023, 5, 31), date(2023, 5, 30)),
        (date(2023, 5, 31), lambda d: d == date(2023, 5, 31), date(2023, 5, 30)),
        (date(2023, 6, 1), lambda d: d == date(2023, 6, 30), date(2023, 6, 29)),
        (date(2023, 6, 29), lambda d: d == date(2023, 6, 30), date(2023, 6, 29)),
        (date(2023, 6, 30), lambda d: d == date(2023, 6, 30), date(2023, 6, 29)),
        (date(2023, 7, 1), lambda d: d == date(2023, 7, 31), date(2023, 7, 30)),
        (date(2023, 7, 30), lambda d: d == date(2023, 7, 31), date(2023, 7, 30)),
        (date(2023, 7, 31), lambda d: d == date(2023, 7, 31), date(2023, 7, 30)),
        (date(2023, 8, 1), lambda d: d == date(2023, 8, 31), date(2023, 8, 30)),
        (date(2023, 8, 30), lambda d: d == date(2023, 8, 31), date(2023, 8, 30)),
        (date(2023, 8, 31), lambda d: d == date(2023, 8, 31), date(2023, 8, 30)),
        (date(2023, 9, 1), lambda d: d == date(2023, 9, 30), date(2023, 9, 29)),
        (date(2023, 9, 29), lambda d: d == date(2023, 9, 30), date(2023, 9, 29)),
        (date(2023, 9, 30), lambda d: d == date(2023, 9, 30), date(2023, 9, 29)),
        (date(2023, 10, 1), lambda d: d == date(2023, 10, 31), date(2023, 10, 30)),
        (date(2023, 10, 30), lambda d: d == date(2023, 10, 31), date(2023, 10, 30)),  # noqa: E501
        (date(2023, 10, 31), lambda d: d == date(2023, 10, 31), date(2023, 10, 30)),  # noqa: E501
        (date(2023, 11, 1), lambda d: d == date(2023, 11, 30), date(2023, 11, 29)),  # noqa: E501
        (date(2023, 11, 29), lambda d: d == date(2023, 11, 30), date(2023, 11, 29)),  # noqa: E501
        (date(2023, 11, 30), lambda d: d == date(2023, 11, 30), date(2023, 11, 29)),  # noqa: E501
        (date(2023, 12, 1), lambda d: d == date(2023, 12, 31), date(2023, 12, 30)),  # noqa: E501
        (date(2023, 12, 30), lambda d: d == date(2023, 12, 31), date(2023, 12, 30)),  # noqa: E501
        (date(2023, 12, 31), lambda d: d == date(2023, 12, 31), date(2023, 12, 30)),  # noqa: E501
    ],
)
def test_get_biz_end_of_month(
    d: date,
    is_holiday: Callable[[date], bool],
    expected: date,
) -> None:
    assert get_biz_end_of_month(d, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,years,months,is_holiday,expected",
    [
        (date(2024, 2, 1), 0, 0, lambda _: False, date(2024, 2, 1)),
        (date(2024, 2, 15), 0, 0, lambda _: False, date(2024, 2, 15)),
        (date(2024, 2, 28), 0, 0, lambda _: False, date(2024, 2, 28)),
        (date(2024, 2, 29), 0, 0, lambda _: False, date(2024, 2, 29)),
        (date(2024, 3, 31), 0, 0, lambda _: False, date(2024, 3, 31)),
        (date(2024, 2, 1), 0, -1, lambda _: False, date(2024, 1, 1)),
        (date(2024, 2, 15), 0, -1, lambda _: False, date(2024, 1, 15)),
        (date(2024, 2, 28), 0, -1, lambda _: False, date(2024, 1, 28)),
        (date(2024, 2, 29), 0, -1, lambda _: False, date(2024, 1, 29)),
        (date(2024, 3, 31), 0, -1, lambda _: False, date(2024, 2, 29)),
        (date(2024, 2, 1), 0, -2, lambda _: False, date(2023, 12, 1)),
        (date(2024, 2, 15), 0, -2, lambda _: False, date(2023, 12, 15)),
        (date(2024, 2, 28), 0, -2, lambda _: False, date(2023, 12, 28)),
        (date(2024, 2, 29), 0, -2, lambda _: False, date(2023, 12, 29)),
        (date(2024, 3, 31), 0, -2, lambda _: False, date(2024, 1, 31)),
    ],
)
def test_add_years_months_with_bizeom2bizeomFalse_bizsom2bizsomFalse(
    d: date,
    years: int,
    months: int,
    is_holiday: Callable[[date], bool],
    expected: date,
) -> None:
    actual = add_years_months(
        d,
        years,
        months,
        is_holiday,
        bizeom2bizeom=False,
        bizsom2bizsom=False,
    )
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,years,months,is_holiday,expected",
    [
        (date(2024, 2, 1), 0, 0, lambda _: False, date(2024, 2, 1)),
        (date(2024, 2, 15), 0, 0, lambda _: False, date(2024, 2, 15)),
        (date(2024, 2, 28), 0, 0, lambda _: False, date(2024, 2, 28)),
        (date(2024, 2, 29), 0, 0, lambda _: False, date(2024, 2, 29)),
        (date(2024, 3, 31), 0, 0, lambda _: False, date(2024, 3, 31)),
        (date(2024, 2, 1), 0, -1, lambda _: False, date(2024, 1, 1)),
        (date(2024, 2, 15), 0, -1, lambda _: False, date(2024, 1, 15)),
        (date(2024, 2, 28), 0, -1, lambda _: False, date(2024, 1, 28)),
        (date(2024, 2, 29), 0, -1, lambda _: False, date(2024, 1, 31)),
        (date(2024, 3, 31), 0, -1, lambda _: False, date(2024, 2, 29)),
        (date(2024, 2, 1), 0, -1, lambda d: d == date(2024, 1, 1), date(2024, 1, 2)),  # noqa: E501
        (date(2024, 2, 15), 0, -1, lambda d: d == date(2024, 1, 15), date(2024, 1, 15)),  # noqa: E501
        (date(2024, 2, 28), 0, -1, lambda d: d == date(2024, 1, 28), date(2024, 1, 28)),  # noqa: E501
        (date(2024, 2, 29), 0, -1, lambda d: d == date(2024, 1, 29), date(2024, 1, 31)),  # noqa: E501
        (date(2024, 3, 31), 0, -1, lambda d: d == date(2024, 2, 29), date(2024, 2, 28)),  # noqa: E501
        (date(2024, 2, 1), 0, 1, lambda _: False, date(2024, 3, 1)),
        (date(2024, 2, 15), 0, 1, lambda _: False, date(2024, 3, 15)),
        (date(2024, 2, 28), 0, 1, lambda _: False, date(2024, 3, 28)),
        (date(2024, 2, 29), 0, 1, lambda _: False, date(2024, 3, 31)),
        (date(2024, 3, 31), 0, 1, lambda _: False, date(2024, 4, 30)),
        (date(2024, 2, 1), 0, 1, lambda d: d == date(2024, 3, 1), date(2024, 3, 2)),  # noqa: E501
        (date(2024, 2, 15), 0, 1, lambda d: d == date(2024, 3, 15), date(2024, 3, 15)),  # noqa: E501
        (date(2024, 2, 28), 0, 1, lambda d: d == date(2024, 3, 28), date(2024, 3, 28)),  # noqa: E501
        (date(2024, 2, 29), 0, 1, lambda d: d == date(2024, 3, 31), date(2024, 3, 30)),  # noqa: E501
        (date(2024, 3, 31), 0, 1, lambda d: d == date(2024, 4, 30), date(2024, 4, 29)),  # noqa: E501
        (date(2024, 2, 1), 1, 0, lambda _: False, date(2025, 2, 1)),
        (date(2024, 2, 15), 1, 0, lambda _: False, date(2025, 2, 15)),
        (date(2024, 2, 28), 1, 0, lambda _: False, date(2025, 2, 28)),
        (date(2024, 2, 29), 1, 0, lambda _: False, date(2025, 2, 28)),
        (date(2024, 3, 31), 1, 0, lambda _: False, date(2025, 3, 31)),
        (date(2024, 2, 1), 1, 0, lambda d: d == date(2025, 2, 1), date(2025, 2, 2)),  # noqa: E501
        (date(2024, 2, 15), 1, 0, lambda d: d == date(2025, 2, 15), date(2025, 2, 15)),  # noqa: E501
        (date(2024, 2, 28), 1, 0, lambda d: d == date(2025, 2, 28), date(2025, 2, 28)),  # noqa: E501
        (date(2024, 2, 29), 1, 0, lambda d: d == date(2025, 2, 28), date(2025, 2, 27)),  # noqa: E501
        (date(2024, 3, 31), 1, 0, lambda d: d == date(2025, 3, 31), date(2025, 3, 30)),  # noqa: E501
        (date(2025, 2, 28), -1, 0, lambda _: False, date(2024, 2, 29)),
    ],
)
def test_add_years_months_with_bizeom2bizeomTrue_bizsom2bizsomTrue(
    d: date,
    years: int,
    months: int,
    is_holiday: Callable[[date], bool],
    expected: date,
) -> None:
    actual = add_years_months(
        d,
        years,
        months,
        is_holiday,
        bizeom2bizeom=True,
        bizsom2bizsom=True,
    )
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, months, is_holiday, bizeom2bizeom, bizsom2bizsom",
    [
        (
            date(2023, 1, 1),
            0,
            lambda _: False,
            True,
            True,
        ),
        (
            date(2023, 1, 5),
            -1,
            lambda _: False,
            True,
            False,
        ),
        (
            date(2023, 1, 5),
            -2,
            lambda _: False,
            False,
            True,
        ),
        (
            date(2023, 1, 5),
            -3,
            lambda _: False,
            False,
            False,
        ),
        (
            date(2023, 1, 5),
            1,
            lambda d: d == date(2023, 1, 25),
            True,
            True,
        ),
        (
            date(2023, 1, 5),
            2,
            lambda d: d == date(2023, 1, 25),
            False,
            True,
        ),
        (
            date(2023, 1, 5),
            3,
            lambda d: d == date(2023, 1, 25),
            False,
            False,
        ),
        (
            date(2023, 1, 5),
            4,
            lambda d: d == date(2023, 1, 25),
            True,
            False,
        ),
    ],
)
def test_add_months(
    date: date,
    months: int,
    is_holiday: Callable[[date], bool],
    bizeom2bizeom: bool,
    bizsom2bizsom: bool,
    mocker: MockerFixture,
) -> None:
    # Spy on the add_years_months function to verify its usage
    add_years_months_spy = mocker.spy(pybizday_utils.month, "add_years_months")
    # Call the add_months function
    _ = add_months(
        date,
        months,
        is_holiday,
        bizeom2bizeom=bizeom2bizeom,
        bizsom2bizsom=bizsom2bizsom,
    )
    # Assert
    add_years_months_spy.assert_called_once_with(
        date,
        0,
        months,
        is_holiday,
        bizeom2bizeom=bizeom2bizeom,
        bizsom2bizsom=bizsom2bizsom,
        datetime_handler=datetime.date,
    )


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, years, is_holiday, bizeom2bizeom, bizsom2bizsom",
    [
        (
            date(2023, 1, 1),
            0,
            lambda _: False,
            True,
            True,
        ),
        (
            date(2023, 1, 5),
            -1,
            lambda _: False,
            True,
            False,
        ),
        (
            date(2023, 1, 5),
            -2,
            lambda _: False,
            False,
            True,
        ),
        (
            date(2023, 1, 5),
            -3,
            lambda _: False,
            False,
            False,
        ),
        (
            date(2023, 1, 5),
            1,
            lambda d: d == date(2023, 1, 25),
            True,
            True,
        ),
        (
            date(2023, 1, 5),
            2,
            lambda d: d == date(2023, 1, 25),
            False,
            True,
        ),
        (
            date(2023, 1, 5),
            3,
            lambda d: d == date(2023, 1, 25),
            False,
            False,
        ),
        (
            date(2023, 1, 5),
            4,
            lambda d: d == date(2023, 1, 25),
            True,
            False,
        ),
    ],
)
def test_add_years(
    date: date,
    years: int,
    is_holiday: Callable[[date], bool],
    bizeom2bizeom: bool,
    bizsom2bizsom: bool,
    mocker: MockerFixture,
) -> None:
    # Spy on the add_years_months function to verify its usage
    add_years_months_spy = mocker.spy(pybizday_utils.month, "add_years_months")
    # Call the add_years function
    _ = add_years(
        date,
        years,
        is_holiday,
        bizeom2bizeom=bizeom2bizeom,
        bizsom2bizsom=bizsom2bizsom,
    )
    # Assert
    add_years_months_spy.assert_called_once_with(
        date,
        years,
        0,
        is_holiday,
        bizeom2bizeom=bizeom2bizeom,
        bizsom2bizsom=bizsom2bizsom,
        datetime_handler=datetime.date,
    )


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,handler",
    [
        (datetime(2023, 1, 1, 12), lambda _: False, datetime.date),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=1)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=2)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=3)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=4)).date(),
        ),
    ],
)
def test_is_biz_end_of_month_with_datetime_obj_should_return_the_same_result_of_is_biz_end_of_month_with_handled_date(  # noqa: E501
    d: datetime,
    is_holiday: Callable[[date], bool],
    handler: Callable[[datetime], date],
) -> None:
    handled = handler(d)
    actual = is_biz_end_of_month(d, is_holiday, datetime_handler=handler)
    expected = is_biz_end_of_month(handled, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,handler",
    [
        (datetime(2023, 1, 1, 12), lambda _: False, datetime.date),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=1)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=2)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=3)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=4)).date(),
        ),
    ],
)
def test_is_biz_start_of_month_with_datetime_obj_should_return_the_same_result_of_is_biz_start_of_month_with_handled_date(  # noqa: E501
    d: datetime,
    is_holiday: Callable[[date], bool],
    handler: Callable[[datetime], date],
) -> None:
    handled = handler(d)
    actual = is_biz_start_of_month(d, is_holiday, datetime_handler=handler)
    expected = is_biz_start_of_month(handled, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,handler",
    [
        (datetime(2023, 1, 1, 12), lambda _: False, datetime.date),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=1)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=2)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=3)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=4)).date(),
        ),
    ],
)
def test_get_biz_end_of_month_with_datetime_obj_should_return_the_same_result_of_get_biz_end_of_month_with_handled_date(  # noqa: E501
    d: datetime,
    is_holiday: Callable[[date], bool],
    handler: Callable[[datetime], date],
) -> None:
    handled = handler(d)
    actual = get_biz_end_of_month(d, is_holiday, datetime_handler=handler)
    expected = get_biz_end_of_month(handled, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,is_holiday,handler",
    [
        (datetime(2023, 1, 1, 12), lambda _: False, datetime.date),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=1)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=2)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=3)).date(),
        ),
        (
            datetime(2023, 1, 1, 12),
            lambda _: False,
            lambda x: (x + timedelta(days=4)).date(),
        ),
    ],
)
def test_get_biz_start_of_month_with_datetime_obj_should_return_the_same_result_of_get_biz_start_of_month_with_handled_date(  # noqa: E501
    d: datetime,
    is_holiday: Callable[[date], bool],
    handler: Callable[[datetime], date],
) -> None:
    handled = handler(d)
    actual = get_biz_start_of_month(d, is_holiday, datetime_handler=handler)
    expected = get_biz_start_of_month(handled, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d,years,months,is_holiday,handler",
    [
        (datetime(2024, 2, 1, 12), 1, 1, lambda _: False, datetime.date),
        (
            datetime(2024, 2, 15, 12),
            1,
            1,
            lambda _: False,
            lambda x: (x + timedelta(days=1)).date(),
        ),
        (
            datetime(2024, 2, 28, 12),
            1,
            1,
            lambda _: False,
            lambda x: (x + timedelta(days=2)).date(),
        ),
        (
            datetime(2024, 2, 29, 12),
            1,
            1,
            lambda _: False,
            lambda x: (x + timedelta(days=3)).date(),
        ),
        (
            datetime(2024, 3, 31, 12),
            1,
            1,
            lambda _: False,
            lambda x: (x + timedelta(days=4)).date(),
        ),
    ],
)
def test_add_years_months_with_datetime_obj_should_return_the_same_result_of_add_years_months_with_handled_date(  # noqa: E501
    d: datetime,
    years: int,
    months: int,
    is_holiday: Callable[[date], bool],
    handler: Callable[[datetime], date],
) -> None:
    handled = handler(d)
    actual = add_years_months(
        d,
        years,
        months,
        is_holiday,
        datetime_handler=handler,
    )
    expected = add_years_months(handled, years, months, is_holiday)
    assert actual == expected



import datetime
from typing import Callable, Generator

from .utils import validate_date_type


def date_range(
    start: datetime.date | datetime.datetime,
    end: datetime.date | datetime.datetime | None = None,
    *,
    include_start: bool = True,
    include_end: bool = True,
    step_days: int = 1,
    datetime_handler: Callable[
        [datetime.datetime], datetime.date
    ] = datetime.datetime.date,
) -> Generator[datetime.date, None, None]:
    """date generator from start to end.

    Args:
        start (datetime.date | datetime.datetime): start date
        end (datetime.date | datetime.datetime | None, optional): end date.
            Defaults to None.
        include_start (bool, optional): include start date. Defaults to True.
        include_end (bool, optional): include end date. Defaults to True.
        step_days (int, optional): step days. Defaults to 1
        datetime_handler (Callable[[datetime.datetime], datetime.date], optional):
            function to convert datetime.datetime to datetime.date.
            Defaults to datetime.datetime.date.

    Yields:
        Generator[datetime.date, None, None]: date generator

    Raises:
        ValueError: step_days is 0
        TypeError: If start or end is not a datetime.date or datetime.datetime.

    Notes:
        - the generator will stop when it reaches out of range, that is,
          when the date is greater than date.max or less than date.min.
    """
    # validate step_days
    if step_days == 0:
        raise ValueError("step_days must not be 0")

    # validate and convert start and end to date
    validate_date_type(start)
    if end is not None:
        validate_date_type(end)

    if isinstance(start, datetime.datetime):
        start = datetime_handler(start)
    if isinstance(end, datetime.datetime):
        end = datetime_handler(end)

    # set ascending and delta
    ASCENDING = step_days > 0
    DELTA = datetime.timedelta(days=step_days)

    # set start date and end date
    if not include_start:
        start += DELTA
    if ASCENDING and end is None:
        end = datetime.date.max
    elif not ASCENDING and end is None:
        end = datetime.date.min
    assert end is not None

    # set end date
    is_broken: Callable[[datetime.date], bool] = {
        (True, True): end.__lt__,
        (True, False): end.__le__,
        (False, True): end.__gt__,
        (False, False): end.__ge__,
    }[(ASCENDING, include_end)]

    # yield date
    try:
        while not is_broken(start):
            yield start
            start += DELTA
    except OverflowError:
        pass



import datetime
from typing import Callable

import pytest

from pybizday_utils.date_range_utils import date_range


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, include_start, include_end, step_days, expected",
    [
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 3),
            True,
            True,
            1,
            [
                datetime.date(2021, 1, 1),
                datetime.date(2021, 1, 2),
                datetime.date(2021, 1, 3),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 3),
            True,
            False,
            1,
            [
                datetime.date(2021, 1, 1),
                datetime.date(2021, 1, 2),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 3),
            False,
            True,
            1,
            [
                datetime.date(2021, 1, 2),
                datetime.date(2021, 1, 3),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 3),
            False,
            False,
            1,
            [
                datetime.date(2021, 1, 2),
            ],
        ),
        (
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 1),
            True,
            True,
            -1,
            [
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 2),
                datetime.date(2021, 1, 1),
            ],
        ),
        (
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 1),
            True,
            False,
            -1,
            [
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 2),
            ],
        ),
        (
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 1),
            False,
            True,
            -1,
            [
                datetime.date(2021, 1, 2),
                datetime.date(2021, 1, 1),
            ],
        ),
        (
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 1),
            False,
            False,
            -1,
            [
                datetime.date(2021, 1, 2),
            ],
        ),
    ],
)
def test_date_range_with_end(
    start: datetime.date,
    end: datetime.date,
    include_start: bool,
    include_end: bool,
    step_days: int,
    expected: list[datetime.date],
) -> None:
    actual = list(
        date_range(
            start,
            end,
            include_start=include_start,
            include_end=include_end,
            step_days=step_days,
        )
    )
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, include_start, include_end, step_days, expected",
    [
        (
            datetime.date.max - datetime.timedelta(days=3),
            True,
            True,
            1,
            [
                datetime.date.max - datetime.timedelta(days=3),
                datetime.date.max - datetime.timedelta(days=2),
                datetime.date.max - datetime.timedelta(days=1),
                datetime.date.max,
            ],
        ),
        (
            datetime.date.max - datetime.timedelta(days=3),
            True,
            False,
            1,
            [
                datetime.date.max - datetime.timedelta(days=3),
                datetime.date.max - datetime.timedelta(days=2),
                datetime.date.max - datetime.timedelta(days=1),
            ],
        ),
        (
            datetime.date.max - datetime.timedelta(days=3),
            False,
            True,
            1,
            [
                datetime.date.max - datetime.timedelta(days=2),
                datetime.date.max - datetime.timedelta(days=1),
                datetime.date.max,
            ],
        ),
        (
            datetime.date.max - datetime.timedelta(days=3),
            False,
            False,
            1,
            [
                datetime.date.max - datetime.timedelta(days=2),
                datetime.date.max - datetime.timedelta(days=1),
            ],
        ),
        (
            datetime.date.min + datetime.timedelta(days=3),
            True,
            True,
            -1,
            [
                datetime.date.min + datetime.timedelta(days=3),
                datetime.date.min + datetime.timedelta(days=2),
                datetime.date.min + datetime.timedelta(days=1),
                datetime.date.min,
            ],
        ),
        (
            datetime.date.min + datetime.timedelta(days=3),
            True,
            False,
            -1,
            [
                datetime.date.min + datetime.timedelta(days=3),
                datetime.date.min + datetime.timedelta(days=2),
                datetime.date.min + datetime.timedelta(days=1),
            ],
        ),
        (
            datetime.date.min + datetime.timedelta(days=3),
            False,
            True,
            -1,
            [
                datetime.date.min + datetime.timedelta(days=2),
                datetime.date.min + datetime.timedelta(days=1),
                datetime.date.min,
            ],
        ),
        (
            datetime.date.min + datetime.timedelta(days=3),
            False,
            False,
            -1,
            [
                datetime.date.min + datetime.timedelta(days=2),
                datetime.date.min + datetime.timedelta(days=1),
            ],
        ),
    ],
)
def test_date_range_without_end(
    start: datetime.date,
    include_start: bool,
    include_end: bool,
    step_days: int,
    expected: list[datetime.date],
) -> None:
    actual = list(
        date_range(
            start,
            include_start=include_start,
            include_end=include_end,
            step_days=step_days,
        )
    )
    assert actual == expected


@pytest.mark.negative
def test_date_range_with_invalid_step_days() -> None:
    with pytest.raises(ValueError):
        list(date_range(datetime.date(2021, 1, 1), step_days=0))


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, include_start, include_end, step_days, expected",
    [
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 7),
            True,
            True,
            2,
            [
                datetime.date(2021, 1, 1),
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 5),
                datetime.date(2021, 1, 7),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 7),
            True,
            False,
            2,
            [
                datetime.date(2021, 1, 1),
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 5),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 7),
            False,
            True,
            2,
            [
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 5),
                datetime.date(2021, 1, 7),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 7),
            False,
            False,
            2,
            [
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 5),
            ],
        ),
        (
            datetime.date(2021, 1, 7),
            datetime.date(2021, 1, 1),
            True,
            True,
            -2,
            [
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 5),
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 1),
            ],
        ),
        (
            datetime.date(2021, 1, 7),
            datetime.date(2021, 1, 1),
            True,
            False,
            -2,
            [
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 5),
                datetime.date(2021, 1, 3),
            ],
        ),
        (
            datetime.date(2021, 1, 7),
            datetime.date(2021, 1, 1),
            False,
            True,
            -2,
            [
                datetime.date(2021, 1, 5),
                datetime.date(2021, 1, 3),
                datetime.date(2021, 1, 1),
            ],
        ),
        (
            datetime.date(2021, 1, 7),
            datetime.date(2021, 1, 1),
            False,
            False,
            -2,
            [
                datetime.date(2021, 1, 5),
                datetime.date(2021, 1, 3),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 13),
            True,
            True,
            3,
            [
                datetime.date(2021, 1, 1),
                datetime.date(2021, 1, 4),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 10),
                datetime.date(2021, 1, 13),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 13),
            True,
            False,
            3,
            [
                datetime.date(2021, 1, 1),
                datetime.date(2021, 1, 4),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 10),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 13),
            False,
            True,
            3,
            [
                datetime.date(2021, 1, 4),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 10),
                datetime.date(2021, 1, 13),
            ],
        ),
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 13),
            False,
            False,
            3,
            [
                datetime.date(2021, 1, 4),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 10),
            ],
        ),
        (
            datetime.date(2021, 1, 13),
            datetime.date(2021, 1, 1),
            True,
            True,
            -3,
            [
                datetime.date(2021, 1, 13),
                datetime.date(2021, 1, 10),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 4),
                datetime.date(2021, 1, 1),
            ],
        ),
        (
            datetime.date(2021, 1, 13),
            datetime.date(2021, 1, 1),
            True,
            False,
            -3,
            [
                datetime.date(2021, 1, 13),
                datetime.date(2021, 1, 10),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 4),
            ],
        ),
        (
            datetime.date(2021, 1, 13),
            datetime.date(2021, 1, 1),
            False,
            True,
            -3,
            [
                datetime.date(2021, 1, 10),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 4),
                datetime.date(2021, 1, 1),
            ],
        ),
        (
            datetime.date(2021, 1, 13),
            datetime.date(2021, 1, 1),
            False,
            False,
            -3,
            [
                datetime.date(2021, 1, 10),
                datetime.date(2021, 1, 7),
                datetime.date(2021, 1, 4),
            ],
        ),
    ],
)
def test_date_range_with_not_1_step_days(
    start: datetime.date,
    end: datetime.date,
    include_start: bool,
    include_end: bool,
    step_days: int,
    expected: list[datetime.date],
) -> None:
    actual = list(
        date_range(
            start,
            end,
            include_start=include_start,
            include_end=include_end,
            step_days=step_days,
        )
    )
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, include_start, include_end, step_days, handler",
    [
        (
            datetime.datetime(2021, 1, 1, 12, 0),
            datetime.datetime(2021, 1, 3, 12, 0),
            True,
            True,
            1,
            lambda dt: dt.date(),
        ),
        (
            datetime.datetime(2021, 1, 1, 12, 0),
            datetime.datetime(2021, 1, 3, 12, 0),
            True,
            False,
            1,
            lambda dt: dt.date(),
        ),
        (
            datetime.datetime(2021, 1, 1, 0, 0),
            datetime.datetime(2021, 1, 3, 0, 0),
            False,
            True,
            1,
            lambda dt: (dt + datetime.timedelta(days=10)).date(),
        ),
    ],
)
def test_date_range_with_datetime_should_return_the_same_result_of_date_range_with_handled_object(  # noqa: E501
    start: datetime.datetime,
    end: datetime.datetime,
    include_start: bool,
    include_end: bool,
    step_days: int,
    handler: Callable[[datetime.datetime], datetime.date],
) -> None:
    handled_start = handler(start)
    handled_end = handler(end)
    actual = list(
        date_range(
            start,
            end,
            include_start=include_start,
            include_end=include_end,
            step_days=step_days,
            datetime_handler=handler,
        )
    )
    expected = list(
        date_range(
            handled_start,
            handled_end,
            include_start=include_start,
            include_end=include_end,
            step_days=step_days,
        )
    )
    assert actual == expected



from datetime import date

import pytest

from pybizday_utils.default_holiday_utils import (
    _GlobalDefaultHolidayDiscriminator,
    add_global_is_holiday_funcs,
    get_global_holiday_funcs,
    get_global_holiday_funcs_names,
    remove_global_is_holiday_funcs,
    with_is_holiday_funcs,
)
from pybizday_utils.holiday_utils import is_saturday_or_sunday


# mock functions for testing
def func1(d: date) -> bool:
    return d.day == 1


def func2(d: date) -> bool:
    return d.day == 2


def func3_(d: date) -> bool:
    return d.day == 3


def func_new(d: date) -> bool:
    return d.day == 4


def func_new2_(d: date) -> bool:
    return d.day == 5


@pytest.mark.positive
def test_global_default_holiday_discriminator_singleton() -> None:
    # Test that the singleton instance is the same across multiple calls
    instance1 = _GlobalDefaultHolidayDiscriminator.get_instance()
    instance2 = _GlobalDefaultHolidayDiscriminator.get_instance()
    assert instance1 is instance2, "Singleton instances are not the same"


@pytest.mark.positive
def test_get_global_holiday_funcs() -> None:
    # Test that the global holiday functions are empty by default
    default = _GlobalDefaultHolidayDiscriminator.get_instance()
    assert get_global_holiday_funcs() == {}, "Global holiday functions should be empty"  # noqa: E501

    default._is_holiday_funcs["test_func"] = lambda date: date.day == 1
    assert get_global_holiday_funcs() == {
        "test_func": default._is_holiday_funcs["test_func"],
    }

    default._is_holiday_funcs["test_func2"] = lambda date: date.day == 2
    assert get_global_holiday_funcs() == {
        "test_func": default._is_holiday_funcs["test_func"],
        "test_func2": default._is_holiday_funcs["test_func2"],
    }


@pytest.mark.positive
def test_get_global_holiday_funcs_names() -> None:
    # Test that the global holiday functions names are empty by default
    default = _GlobalDefaultHolidayDiscriminator.get_instance()
    assert get_global_holiday_funcs_names() == [], (
        "Global holiday functions names should be empty"
    )

    default._is_holiday_funcs["test_func"] = lambda date: date.day == 1
    assert get_global_holiday_funcs_names() == ["test_func"]

    default._is_holiday_funcs["test_func2"] = lambda date: date.day == 2
    assert get_global_holiday_funcs_names() == ["test_func", "test_func2"]


@pytest.mark.positive
def test_add_global_is_holiday_funcs() -> None:
    # execute
    add_global_is_holiday_funcs(func1, func2, func3=func3_)

    # check
    # NOTE: Assume that get_global_holiday_funcs works as expected.
    #       See test_get_global_holiday_funcs for the test.
    assert get_global_holiday_funcs()["func1"] == func1
    assert get_global_holiday_funcs()["func2"] == func2
    assert get_global_holiday_funcs()["func3"] == func3_


@pytest.mark.positive
def test_add_global_is_holiday_funcs_with_allow_overwrite() -> None:
    # preparation
    add_global_is_holiday_funcs(func1, func2, func3=func3_)

    # execute
    add_global_is_holiday_funcs(func1=func_new, allow_overwrite=True)

    # check
    # NOTE: Assume that get_global_holiday_funcs works as expected.
    #       See test_get_global_holiday_funcs for the test.
    assert get_global_holiday_funcs()["func1"] is func_new


@pytest.mark.negative
def test_add_global_is_holiday_funcs_should_prohibit_overwrite_in_default() -> None:  # noqa: E501
    # execute
    add_global_is_holiday_funcs(func1, func2, func3=func3_)

    # check
    with pytest.raises(ValueError):
        add_global_is_holiday_funcs(func1)
    with pytest.raises(ValueError):
        add_global_is_holiday_funcs(func2)
    with pytest.raises(ValueError):
        add_global_is_holiday_funcs(func3=func3_)


@pytest.mark.positive
def test_remove_global_is_holiday_funcs() -> None:
    # preparation
    add_global_is_holiday_funcs(func1, func2, func3=func3_)

    # execute
    remove_global_is_holiday_funcs("func1", "func3")

    # check
    # NOTE: Assume that get_global_holiday_funcs works as expected.
    #       See test_get_global_holiday_funcs for the test.
    assert "func1" not in get_global_holiday_funcs()
    assert "func3" not in get_global_holiday_funcs()
    assert "func2" in get_global_holiday_funcs()
    assert get_global_holiday_funcs()["func2"] is func2


@pytest.mark.negative
def test_remove_global_is_holiday_funcs_with_non_existing_name() -> None:
    # Test that removing a non-existing function name does not raise an error
    with pytest.raises(KeyError):
        remove_global_is_holiday_funcs("non_existing_func")


@pytest.mark.positive
def test_with_is_holiday_funcs_in_default() -> None:
    # preparation
    add_global_is_holiday_funcs(func1, func2, func3=func3_)
    assert get_global_holiday_funcs()["func1"] is func1
    assert get_global_holiday_funcs()["func2"] is func2
    assert get_global_holiday_funcs()["func3"] is func3_
    assert set(get_global_holiday_funcs_names()) == {"func1", "func2", "func3"}

    # execute and check
    with with_is_holiday_funcs(func_new):
        assert get_global_holiday_funcs()["func1"] is func1
        assert get_global_holiday_funcs()["func2"] is func2
        assert get_global_holiday_funcs()["func3"] is func3_
        assert get_global_holiday_funcs()["func_new"] is func_new
        assert set(get_global_holiday_funcs_names()) == {
            "func1",
            "func2",
            "func3",
            "func_new",
        }  # noqa: E501

    # check after exiting the context manager
    assert get_global_holiday_funcs()["func1"] is func1
    assert get_global_holiday_funcs()["func2"] is func2
    assert get_global_holiday_funcs()["func3"] is func3_
    assert "func_new" not in get_global_holiday_funcs()
    assert set(get_global_holiday_funcs_names()) == {"func1", "func2", "func3"}


@pytest.mark.positive
def test_with_is_holiday_funcs_with_allow_overwrite() -> None:
    # preparation
    add_global_is_holiday_funcs(func1, func2, func3=func3_)
    assert get_global_holiday_funcs()["func1"] is func1
    assert get_global_holiday_funcs()["func2"] is func2
    assert get_global_holiday_funcs()["func3"] is func3_
    assert set(get_global_holiday_funcs_names()) == {"func1", "func2", "func3"}  # noqa: E501

    # execute and check
    with with_is_holiday_funcs(func1=func_new, allow_overwrite=True):
        # check whether func1 is replaced by func_new
        assert get_global_holiday_funcs()["func1"] is func_new
        assert get_global_holiday_funcs()["func2"] is func2
        assert get_global_holiday_funcs()["func3"] is func3_
        assert set(get_global_holiday_funcs_names()) == {"func1", "func2", "func3"}  # noqa: E501

    # check after exiting the context manager
    # check whether func1 is restored to its original value
    assert get_global_holiday_funcs()["func1"] is func1
    assert get_global_holiday_funcs()["func2"] is func2
    assert get_global_holiday_funcs()["func3"] is func3_
    assert set(get_global_holiday_funcs_names()) == {"func1", "func2", "func3"}  # noqa: E501


@pytest.mark.positive
def test_with_is_holiday_funcs_with_all_replace() -> None:
    # preparation
    add_global_is_holiday_funcs(func1, func2, func3=func3_)
    assert get_global_holiday_funcs()["func1"] is func1
    assert get_global_holiday_funcs()["func2"] is func2
    assert get_global_holiday_funcs()["func3"] is func3_
    assert set(get_global_holiday_funcs_names()) == {"func1", "func2", "func3"}  # noqa: E501

    # execute and check
    with with_is_holiday_funcs(func_new, func_new2=func_new2_, all_replace=True):  # noqa: E501
        assert get_global_holiday_funcs()["func_new"] is func_new
        assert get_global_holiday_funcs()["func_new2"] is func_new2_
        assert "func1" not in get_global_holiday_funcs()
        assert "func2" not in get_global_holiday_funcs()
        assert "func3" not in get_global_holiday_funcs()
        assert set(get_global_holiday_funcs_names()) == {"func_new", "func_new2"}  # noqa: E501

    # check after exiting the context manager
    assert get_global_holiday_funcs()["func1"] is func1
    assert get_global_holiday_funcs()["func2"] is func2
    assert get_global_holiday_funcs()["func3"] is func3_
    assert "func_new" not in get_global_holiday_funcs()
    assert "func_new2" not in get_global_holiday_funcs()
    assert set(get_global_holiday_funcs_names()) == {"func1", "func2", "func3"}  # noqa: E501


@pytest.mark.positive
@pytest.mark.use_global_default_holiday_discriminator
def test_default_holiday_discriminator_status() -> None:
    # pereparation
    default = _GlobalDefaultHolidayDiscriminator.get_instance()
    name = is_saturday_or_sunday.__name__
    expected = is_saturday_or_sunday
    # check
    assert default.is_holiday_funcs[name] is expected
    assert get_global_holiday_funcs_names() == [name]
    assert get_global_holiday_funcs()[name] is expected



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

from .holiday_utils import (
    HolidayDiscriminator,
    IsHolidayFuncType,
    is_saturday_or_sunday,
)


class _GlobalDefaultHolidayDiscriminator(HolidayDiscriminator):
    """Singleton class to manage the global default holiday functions.
    This class is a singleton that holds the default holiday functions.

    See Also:
        - HolidayDiscriminator: The base class for managing holiday functions.
    """

    _instance: _GlobalDefaultHolidayDiscriminator | None = None

    def __new__(cls, *args, **kwargs) -> _GlobalDefaultHolidayDiscriminator:  # type: ignore  # noqa: E501
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    @classmethod
    def get_instance(cls) -> _GlobalDefaultHolidayDiscriminator:
        """Get the singleton instance of the class.

        Returns:
            _GlobalDefaultHolidayDiscriminator: The singleton instance of the class.
        """  # noqa: E501
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance


def get_global_holiday_funcs_names() -> list[str]:
    """Get the names of the registered holiday functions.

    Returns:
        list[str]: List of names of the registered holiday functions in the global default holiday discriminator.
    """  # noqa: E501
    default = _GlobalDefaultHolidayDiscriminator.get_instance()
    return default.names


def get_global_holiday_funcs() -> dict[str, IsHolidayFuncType]:
    """Get the registered holiday functions.

    Returns:
        dict[str, IsHolidayFuncType]: Dictionary of the registered holiday functions in the global default holiday discriminator.

    Notes:
        The returned dictionary is a shallow copy of the original dictionary.
        Modifications to the returned dictionary will not affect the original dictionary.
    """  # noqa: E501
    default = _GlobalDefaultHolidayDiscriminator.get_instance()
    return default.is_holiday_funcs


@contextmanager
def with_is_holiday_funcs(
    *is_holiday_funcs_args: IsHolidayFuncType,
    allow_overwrite: bool = False,
    all_replace: bool = False,
    **is_holiday_funcs_kwargs: IsHolidayFuncType,
) -> Generator[HolidayDiscriminator, None, None]:
    """Context manager to add custom holiday functions.
    This context manager allows you to customize the holiday functions used in the global default holiday discriminator temporarily.

    Args:
        *is_holiday_funcs_args (IsHolidayFuncType): Custom holiday functions to add.
        allow_overwrite (bool, optional): Whether to allow overwriting existing holiday functions. Defaults to False.
        all_replace (bool, optional): Whether to remove all current holiday functions before adding new ones. Defaults to False.
        **is_holiday_funcs_kwargs (IsHolidayFuncType): Custom holiday functions to add.

    Yields:
        Generator[HolidayDiscriminator, None, None]: The global default holiday discriminator with the added holiday functions.

    Notes:
        - The context manager will restore the original state of the holiday functions after exiting the context.
        - If `all_replace` is set to True, all current holiday functions will be removed before adding new ones.
        - If `allow_overwrite` is set to True, existing holiday functions with the same name will be overwritten.
    """  # noqa: E501

    # Get the default holiday discriminator instance
    default = _GlobalDefaultHolidayDiscriminator.get_instance()
    # Cache the current holiday functions
    _cache = default.is_holiday_funcs

    # If all_replace is True, remove all current holiday functions
    if all_replace:
        default.remove_is_holiday_funcs(*default.names)

    # add the new holiday functions
    default.add_is_holiday_funcs(
        *is_holiday_funcs_args,
        allow_overwrite=allow_overwrite,
        **is_holiday_funcs_kwargs,
    )

    try:
        yield default
    finally:
        # remove all holiday functions
        default.remove_is_holiday_funcs(*default.names)
        # Restore the original state
        default.add_is_holiday_funcs(**_cache, allow_overwrite=True)


# initialize the default holiday functions
global_default_holiday_discriminator = _GlobalDefaultHolidayDiscriminator.get_instance()  # noqa: E501
global_default_holiday_discriminator.add_is_holiday_funcs(
    is_saturday_or_sunday,
    allow_overwrite=True,
)
add_global_is_holiday_funcs = global_default_holiday_discriminator.add_is_holiday_funcs  # noqa: E501
remove_global_is_holiday_funcs = (
    global_default_holiday_discriminator.remove_is_holiday_funcs
)



import datetime
from typing import Any, TypeGuard


def validate_date_type(date: Any) -> TypeGuard[datetime.date | datetime.datetime]:
    """Validate that the input is either a datetime.date or datetime.datetime object.

    Args:
        date (Any): The value to validate.

    Returns:
        TypeGuard[datetime.date | datetime.datetime]: True for valid date types.

    Raises:
        TypeError: If the input is not a datetime.date or datetime.datetime object.
    """
    if not isinstance(date, (datetime.date, datetime.datetime)):
        raise TypeError(
            f"date must be a datetime.date or datetime.datetime object, "
            f"not {type(date)}"
        )
    return True



import datetime
from typing import Callable

import pytest

from pybizday_utils.basic import (
    bizday_range,
    count_bizdays,
    get_n_next_bizday,
    get_n_prev_bizday,
    get_next_bizday,
    get_prev_bizday,
    is_bizday,
)


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, is_holiday, expected",
    [
        (datetime.date(2021, 1, 1), lambda d: d == datetime.date(2021, 1, 1), False),  # noqa
        (datetime.date(2021, 1, 2), lambda d: d == datetime.date(2021, 1, 3), True),  # noqa
    ],
)
def test_is_bizday(
    date: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    expected: bool,
) -> None:
    assert is_bizday(date, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            lambda _: False,
            datetime.date(2025, 4, 26),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 28),
        ),
    ],
)
def test_get_next_bizday(
    date: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_next_bizday(date, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            lambda _: False,
            datetime.date(2025, 4, 24),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d == (datetime.date(2025, 4, 24)),
            datetime.date(2025, 4, 23),
        ),
        (
            datetime.date(2025, 4, 25),
            lambda d: d in {datetime.date(2025, 4, 24), datetime.date(2025, 4, 23)},  # noqa
            datetime.date(2025, 4, 22),
        ),
    ],
)
def test_get_prev_bizday(
    date: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_prev_bizday(date, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, n, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            1,
            lambda _: False,
            datetime.date(2025, 4, 26),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda _: False,
            datetime.date(2025, 4, 24),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d == (datetime.date(2025, 4, 24)),
            datetime.date(2025, 4, 23),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d in {datetime.date(2025, 4, 24), datetime.date(2025, 4, 23)},  # noqa
            datetime.date(2025, 4, 22),
        ),
        (
            datetime.date(2025, 4, 25),
            0,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda _: False,
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 29),
        ),
        (
            datetime.date(2025, 4, 25),
            2,
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            datetime.date(2025, 4, 30),
        ),
        (
            datetime.date(2025, 4, 27),
            -2,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 28),
            -2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 29),
            -2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 30),
            -2,
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            datetime.date(2025, 4, 25),
        ),
    ],
)
def test_get_n_next_bizday(
    date: datetime.date,
    n: int,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_n_next_bizday(date, n, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.heavy
def test_get_n_next_bizday_with_large_n() -> None:
    date = datetime.date(2025, 4, 25)
    n = 10000
    actual = get_n_next_bizday(date, n, is_holiday=lambda _: False)
    expected = date + datetime.timedelta(days=n)
    assert actual == expected


@pytest.mark.negative
def test_get_n_next_bizday_with_n_0_and_date_holiday() -> None:
    date = datetime.date(2025, 4, 25)

    with pytest.raises(ValueError):
        get_n_next_bizday(date, 0, is_holiday=lambda _: True)


@pytest.mark.positive
@pytest.mark.parametrize(
    "date, n, is_holiday, expected",
    [
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda _: False,
            datetime.date(2025, 4, 26),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            -1,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda _: False,
            datetime.date(2025, 4, 24),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d == (datetime.date(2025, 4, 24)),
            datetime.date(2025, 4, 23),
        ),
        (
            datetime.date(2025, 4, 25),
            1,
            lambda d: d in {datetime.date(2025, 4, 24), datetime.date(2025, 4, 23)},  # noqa
            datetime.date(2025, 4, 22),
        ),
        (
            datetime.date(2025, 4, 25),
            0,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda _: False,
            datetime.date(2025, 4, 27),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 28),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 29),
        ),
        (
            datetime.date(2025, 4, 25),
            -2,
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            datetime.date(2025, 4, 30),
        ),
        (
            datetime.date(2025, 4, 27),
            2,
            lambda _: False,
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 28),
            2,
            lambda d: d == (datetime.date(2025, 4, 26)),
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 29),
            2,
            lambda d: d in {datetime.date(2025, 4, 26), datetime.date(2025, 4, 27)},  # noqa
            datetime.date(2025, 4, 25),
        ),
        (
            datetime.date(2025, 4, 30),
            2,
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            datetime.date(2025, 4, 25),
        ),
    ],
)
def test_get_n_prev_bizday(
    date: datetime.date,
    n: int,
    is_holiday: Callable[[datetime.date], bool],
    expected: datetime.date,
) -> None:
    assert get_n_prev_bizday(date, n, is_holiday) == expected


@pytest.mark.positive
@pytest.mark.heavy
def test_get_n_prev_bizday_with_large_n() -> None:
    date = datetime.date(2025, 4, 25)
    n = 10000
    actual = get_n_prev_bizday(date, n, is_holiday=lambda _: False)
    expected = date - datetime.timedelta(days=n)
    assert actual == expected


@pytest.mark.negative
def test_get_n_prev_bizday_with_n_0_and_date_holiday() -> None:
    date = datetime.date(2025, 4, 25)

    with pytest.raises(ValueError):
        get_n_prev_bizday(date, 0, is_holiday=lambda _: True)


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, is_holiday, include_start, include_end, expected",
    [
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda _: False,
            True,
            True,
            [
                datetime.date(2025, 4, 25),
                datetime.date(2025, 4, 26),
                datetime.date(2025, 4, 27),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 29),
                datetime.date(2025, 4, 30),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            True,
            [
                datetime.date(2025, 4, 25),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 30),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            True,
            [
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 30),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            False,
            [
                datetime.date(2025, 4, 25),
                datetime.date(2025, 4, 28),
            ],
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            False,
            [
                datetime.date(2025, 4, 28),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda _: False,
            True,
            True,
            [
                datetime.date(2025, 4, 30),
                datetime.date(2025, 4, 29),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 27),
                datetime.date(2025, 4, 26),
                datetime.date(2025, 4, 25),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            True,
            [
                datetime.date(2025, 4, 30),
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 25),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            True,
            [
                datetime.date(2025, 4, 28),
                datetime.date(2025, 4, 25),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            False,
            [
                datetime.date(2025, 4, 30),
                datetime.date(2025, 4, 28),
            ],
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            False,
            [
                datetime.date(2025, 4, 28),
            ],
        ),
    ],
)
def test_bizday_range(
    start: datetime.date,
    end: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    include_start: bool,
    include_end: bool,
    expected: list[datetime.date],
) -> None:
    actual = list(
        bizday_range(
            start,
            end,
            is_holiday,
            include_start=include_start,
            include_end=include_end,
        )
    )
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, is_holiday, include_start, include_end, expected",
    [
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda _: False,
            True,
            True,
            6,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            True,
            3,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            True,
            2,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            False,
            2,
        ),
        (
            datetime.date(2025, 4, 25),
            datetime.date(2025, 4, 30),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            False,
            1,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda _: False,
            True,
            True,
            -6,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            True,
            -3,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            True,
            -2,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            True,
            False,
            -2,
        ),
        (
            datetime.date(2025, 4, 30),
            datetime.date(2025, 4, 25),
            lambda d: (
                d
                in {
                    datetime.date(2025, 4, 26),
                    datetime.date(2025, 4, 27),
                    datetime.date(2025, 4, 29),
                }
            ),
            False,
            False,
            -1,
        ),
    ],
)
def test_count_bizdays(
    start: datetime.date,
    end: datetime.date,
    is_holiday: Callable[[datetime.date], bool],
    include_start: bool,
    include_end: bool,
    expected: int,
) -> None:
    actual = count_bizdays(
        start,
        end,
        is_holiday,
        include_start=include_start,
        include_end=include_end,
    )
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, handler, is_holiday",
    [
        (
            datetime.datetime(2021, 1, 1),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 1),
        ),
        (
            datetime.datetime(2021, 1, 2),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 3),
            lambda _: datetime.date(2024, 1, 1),
            lambda d: d == datetime.date(2024, 1, 1),
        ),
    ],
)
def test_is_bizday_with_datetime_obj_should_return_the_same_result_of_is_bizyday_with_handled_obj(  # noqa: E501
    d: datetime.datetime,
    handler: Callable[[datetime.datetime], datetime.date],
    is_holiday: Callable[[datetime.date], bool],
) -> None:
    handled_d = handler(d)
    actual = is_bizday(d, is_holiday, datetime_handler=handler)
    expected = is_bizday(handled_d, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, handler, is_holiday",
    [
        (
            datetime.datetime(2021, 1, 1),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 1),
        ),
        (
            datetime.datetime(2021, 1, 2),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 3),
            lambda _: datetime.date(2024, 1, 3),
            lambda d: d == datetime.date(2024, 1, 4),
        ),
    ],
)
def test_get_next_bizday_with_datetime_obj_should_return_the_same_result_of_get_next_bizday_with_handled_obj(  # noqa: E501
    d: datetime.datetime,
    handler: Callable[[datetime.datetime], datetime.date],
    is_holiday: Callable[[datetime.date], bool],
) -> None:
    handled_d = handler(d)
    actual = get_next_bizday(d, is_holiday, datetime_handler=handler)
    expected = get_next_bizday(handled_d, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, handler, is_holiday",
    [
        (
            datetime.datetime(2021, 1, 1),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 1),
        ),
        (
            datetime.datetime(2021, 1, 2),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 3),
            lambda _: datetime.date(2024, 1, 3),
            lambda d: d == datetime.date(2024, 1, 2),
        ),
    ],
)
def test_get_prev_bizday_with_datetime_obj_should_return_the_same_result_of_get_prev_bizday_with_handled_obj(  # noqa: E501
    d: datetime.datetime,
    handler: Callable[[datetime.datetime], datetime.date],
    is_holiday: Callable[[datetime.date], bool],
) -> None:
    handled_d = handler(d)
    actual = get_prev_bizday(d, is_holiday, datetime_handler=handler)
    expected = get_prev_bizday(handled_d, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, handler, n, is_holiday",
    [
        (
            datetime.datetime(2021, 1, 1),
            lambda dt: dt.date(),
            1,
            lambda d: d == datetime.date(2021, 1, 1),
        ),
        (
            datetime.datetime(2021, 1, 2),
            lambda dt: dt.date(),
            -1,
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 3),
            lambda _: datetime.date(2024, 1, 3),
            2,
            lambda d: d == datetime.date(2024, 1, 4),
        ),
    ],
)
def test_get_n_next_bizday_with_datetime_obj_should_return_the_same_result_of_get_n_next_bizday_with_handled_obj(  # noqa: E501
    d: datetime.datetime,
    handler: Callable[[datetime.datetime], datetime.date],
    n: int,
    is_holiday: Callable[[datetime.date], bool],
) -> None:
    handled_d = handler(d)
    actual = get_n_next_bizday(d, n, is_holiday, datetime_handler=handler)
    expected = get_n_next_bizday(handled_d, n, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "d, handler, n, is_holiday",
    [
        (
            datetime.datetime(2021, 1, 1),
            lambda dt: dt.date(),
            -1,
            lambda d: d == datetime.date(2021, 1, 1),
        ),
        (
            datetime.datetime(2021, 1, 2),
            lambda dt: dt.date(),
            1,
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 3),
            lambda _: datetime.date(2024, 1, 3),
            2,
            lambda d: d == datetime.date(2024, 1, 2),
        ),
    ],
)
def test_get_n_prev_bizday_with_datetime_obj_should_return_the_same_result_of_get_n_prev_bizday_with_handled_obj(  # noqa: E501
    d: datetime.datetime,
    handler: Callable[[datetime.datetime], datetime.date],
    n: int,
    is_holiday: Callable[[datetime.date], bool],
) -> None:
    handled_d = handler(d)
    actual = get_n_prev_bizday(d, n, is_holiday, datetime_handler=handler)
    expected = get_n_prev_bizday(handled_d, n, is_holiday)
    assert actual == expected


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, handler, is_holiday",
    [
        (
            datetime.datetime(2021, 1, 1),
            datetime.datetime(2021, 1, 10),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 2),
            datetime.datetime(2021, 1, 11),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 3),
            datetime.datetime(2021, 1, 12),
            lambda dt: (dt + datetime.timedelta(days=10)).date(),
            lambda d: d == datetime.date(2024, 1, 15),
        ),
    ],
)
def test_bizday_range_with_datetime_obj_should_return_the_same_result_of_bizday_range_with_handled_obj(  # noqa: E501
    start: datetime.datetime,
    end: datetime.datetime,
    handler: Callable[[datetime.datetime], datetime.date],
    is_holiday: Callable[[datetime.date], bool],
) -> None:
    handled_start = handler(start)
    handled_end = handler(end)
    actual = bizday_range(
        start,
        end,
        is_holiday,
        datetime_handler=handler,
    )
    expected = bizday_range(handled_start, handled_end, is_holiday)
    assert list(actual) == list(expected)


@pytest.mark.positive
@pytest.mark.parametrize(
    "start, end, handler, is_holiday",
    [
        (
            datetime.datetime(2021, 1, 1),
            datetime.datetime(2021, 1, 10),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 1),
        ),
        (
            datetime.datetime(2021, 1, 2),
            datetime.datetime(2021, 1, 11),
            lambda dt: dt.date(),
            lambda d: d == datetime.date(2021, 1, 3),
        ),
        (
            datetime.datetime(2021, 1, 3),
            datetime.datetime(2021, 1, 12),
            lambda dt: (dt + datetime.timedelta(days=10)).date(),
            lambda d: d == datetime.date(2024, 1, 15),
        ),
    ],
)
def test_count_bizdays_with_datetime_obj_should_return_the_same_result_of_count_bizdays_with_handled_obj(  # noqa: E501
    start: datetime.datetime,
    end: datetime.datetime,
    handler: Callable[[datetime.datetime], datetime.date],
    is_holiday: Callable[[datetime.date], bool],
) -> None:
    handled_start = handler(start)
    handled_end = handler(end)
    actual = count_bizdays(start, end, is_holiday, datetime_handler=handler)
    expected = count_bizdays(handled_start, handled_end, is_holiday)
    assert actual == expected


@pytest.mark.negative
def test_get_next_bizday_should_raise_ValueError_if_no_next_bizday_found() -> None:
    date = datetime.date.max - datetime.timedelta(days=10)
    with pytest.raises(ValueError, match="No next business day found"):
        get_next_bizday(date, is_holiday=lambda _: True)


@pytest.mark.negative
def test_get_prev_bizday_should_raise_ValueError_if_no_prev_bizday_found() -> None:
    date = datetime.date.min + datetime.timedelta(days=10)
    with pytest.raises(ValueError, match="No previous business day found"):
        get_prev_bizday(date, is_holiday=lambda _: True)


@pytest.mark.negative
@pytest.mark.parametrize(
    "d, n, expected_msg",
    [
        (
            datetime.date.max - datetime.timedelta(days=10),
            1,
            "No 1-th next business day found",
        ),
        (
            datetime.date.max - datetime.timedelta(days=10),
            2,
            "No 2-th next business day found",
        ),
        (
            datetime.date.min + datetime.timedelta(days=10),
            -1,
            "No 1-th previous business day found",
        ),
    ],
)
def test_get_n_next_bizday_should_raise_ValueError_if_no_next_bizday_found(
    d: datetime.date,
    n: int,
    expected_msg: str,
) -> None:
    with pytest.raises(ValueError, match=expected_msg):
        get_n_next_bizday(d, n, is_holiday=lambda _: True)


@pytest.mark.negative
@pytest.mark.parametrize(
    "d, n, expected_msg",
    [
        (
            datetime.date.min + datetime.timedelta(days=10),
            1,
            "No 1-th previous business day found",
        ),
        (
            datetime.date.min + datetime.timedelta(days=10),
            2,
            "No 2-th previous business day found",
        ),
        (
            datetime.date.max - datetime.timedelta(days=10),
            -1,
            "No 1-th next business day found",
        ),
    ],
)
def test_get_n_prev_bizday_should_raise_ValueError_if_no_prev_bizday_found(
    d: datetime.date,
    n: int,
    expected_msg: str,
) -> None:
    with pytest.raises(ValueError, match=expected_msg):
        get_n_prev_bizday(d, n, is_holiday=lambda _: True)



import nox


@nox.session(venv_backend="uv", python=["3.10", "3.11", "3.12", "3.13"])
def test(session):
    session.install("-e", ".[dev]")
    session.run("pytest", *session.posargs)


@nox.session(venv_backend="uv", python=["3.10", "3.11", "3.12", "3.13"])
def lint(session):
    session.install("-e", ".[dev]")
    session.run("ruff", "check", *session.posargs)


@nox.session(venv_backend="uv", python=["3.10", "3.11", "3.12", "3.13"])
def mypy(session):
    session.install("-e", ".[dev]")
    session.run("mypy", "src", "tests")
